/*	pz.c

	The following Pan/Zoom program was written as a simple
	demonstration of use of VWLib and is distributed as a
	part of that public domain library.  This is the same
	program listed on the reference card, except that, for
	space reasons, the version on the reference card omits
	the command line usage printout and comments.
*/

#include <stdio.h>
#include <AFAPI/vwsup.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


int
main(register int argc,
register char **argv)
{
	vwfbox destarea = { {0.0, 0.0},
			    {1.0, 1.0} };
	vwinfoMouse mouse;
	vwbuf image;
	register double mag = 1.0;

	/* Command line ok? */
	if (argc < 2) {
		fprintf(stderr,
			"Usage: %s [image_file_name_format.ppm]\n",
			argv[0]);
		exit(1);
	}

	/* The standard VWLib check-in */
	vwinit();

	/* Turn-on the panel display */
	vwbegin();

	/* Turn-on the mouse */
	vwbeginMouse();

	/* Map the image file */
	image = vwmapFile(argv[1]);

	for (;;) {
		vwfxy t;

		/* Clear the panel buffer to black;
		   this is really needed only for mag
		   less than full screen...
		*/
		vwclear(0);

		/* Get mouse coordinates and compute a
		   "flipped" center coordinates so that
		   panning will allow the image to just
		   leave the wall in any direction.
		*/
		mouse = vwreadMouse(0);
		t.x = 1.0 - mouse.pos.x;
		t.y = 1.0 - mouse.pos.y;

		/* Adjust magnification */
		if (mouse.left) mag *= 0.95;
		if (mouse.center) {
			/* Video wall exit sequence */
			vwexit();
			exit(0);
		}
		if (mouse.right) mag *= 1.05;

		/* Put the image in the panel buffer,
		   with aspect-preserving magnification
		*/
		vwmag(0, destarea, mouse.pos,
		      &image, t, mag, 255);

		p_wait();
		/* Actually update the wall display */
		vwpaint(0);
	}
}
