/*	valid.c

	Program to validate the AFAPI.

	This program simply runs every routine with random data derived
	from a seed given on the command line, and sends the output to
	standard out.  Thus, validation is accomplished by:

	1. Compile and run this code for a known-working AFAPI (such as
	   SHMAPERS), writing the output to a file.

	2. Compile and run this code for the AFAPI under test, writing the
	   output to a file.

	3. Use unix "diff" to compare the outputs.

	1997 by Hank Dietz
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <AFAPI/afapi.h>
#include <AFAPI/inline.h>

#include "rand.h"


/*	Value print helper routines...

	p_bar type is printed in binary; all others are printed using a
	single ordinary printf().
*/

void
print_bar(register p_bar x)
{
	register int i;

	for (i=p_bar_bits; i>0; --i) {
		if ((((p_bar) 1) << (i-1)) & x) {
			putchar('1');
		} else {
			putchar('0');
		}
	}
}

void print_1u(register p_1u x) { printf("%u", x); }
void print_8(register p_8 x) { printf("%d", x); }
void print_8u(register p_8u x) { printf("%u", x); }
void print_16(register p_16 x) { printf("%d", x); }
void print_16u(register p_16u x) { printf("%u", x); }
void print_32(register p_32 x) { printf("%d", x); }
void print_32u(register p_32u x) { printf("%u", x); }
void print_64(register p_64 x) { printf("%lld", x); }
void print_64u(register p_64u x) { printf("%llu", x); }
void print_32f(register p_32f x) { printf("%g", x); }
void print_64f(register p_64f x) { printf("%g", x); }
void print_ptr(register p_ptr x) { printf("%p", x); }

#define	PRINT_ARRAY(suffix, base, count) \
{ \
	register int i; \
	printf("{"); \
	for (i=0; i<((count)-1); ++i) { \
		print_##suffix(*((base) + i)); \
		printf(", "); \
	} \
	print_##suffix(*((base) + ((count) - 1))); \
	printf("}"); \
}


/*	Declare validation data buffers
*/
#define	P_DECLARE(suffix) \
p_##suffix buf_##suffix;

P_ALPHA

#undef	P_DECLARE


/*	Macro for functions of a single argument...
*/
#define	ONEARG(retsuf, name, argsuf) \
{ \
	register p_##retsuf r = 0; \
	register p_##argsuf a = p_randUniq##argsuf(); \
	a = TWEAK(a); \
	r = name(a); \
	{ \
		register int i; \
		for (i=0; i<NPROC; ++i) { \
			printf("PE%d:\t", i); \
			if (i == IPROC) { \
				p_bcastPut##retsuf(r); \
				print_##retsuf(r); \
			} else { \
				print_##retsuf(p_bcastGet##retsuf()); \
			} \
			printf(" = " #name "("); \
			if (i == IPROC) { \
				p_bcastPut##argsuf(a); \
				print_##argsuf(a); \
			} else { \
				print_##argsuf(p_bcastGet##argsuf()); \
			} \
			printf(")\n"); \
		} \
	} \
	putchar('\n'); \
}


/*	Macro for put-get paired functions...
*/
#define	PUTGET(retsuf, name) \
{ \
	register int sender = (p_randSame16u() % NPROC); \
	register p_##retsuf r = 0; \
	if (sender == IPROC) { \
		r = p_randUniq##retsuf(); \
		r = name##Put##retsuf(r); \
	} else { \
		r = name##Get##retsuf(); \
	} \
	{ \
		p_##retsuf t[NPROC]; \
		register int i; \
		for (i=0; i<NPROC; ++i) { \
			if (i == IPROC) { \
				t[i] = r; \
				p_bcastPut##retsuf(r); \
			} else { \
				t[i] = p_bcastGet##retsuf(); \
			} \
		} \
		for (i=0; i<NPROC; ++i) { \
			printf("PE%d:\t", i); \
			if (i == sender) { \
				print_##retsuf(t[i]); \
				printf(" = " #name "Put" #retsuf "("); \
				print_##retsuf(t[i]); \
			} else { \
				print_##retsuf(t[i]); \
				printf(" = " #name "Get" #retsuf "("); \
			} \
			printf(")\n"); \
		} \
	} \
	putchar('\n'); \
}


void
bcastz(void)
{
	register int sender = (p_randSame16u() % NPROC);
	register p_32 len = (p_randSame16u() % 15) + 1;
	register int i, j;
	p_8u r[16], t[16];

	for (i=0; i<len; ++i) {
		r[i] = p_randUniq8u();
	}

	if (sender == IPROC) {
		p_bcastPutz(&(r[0]), len);
	} else {
		p_bcastGetz(&(r[0]), len);
	}

	for (i=0; i<NPROC; ++i) {
		if (i == IPROC) {
			p_bcastPutz(&(r[0]), len);
			for (j=0; j<len; ++j) t[j] = r[j];
		} else {
			p_bcastGetz(&(t[0]), len);
		}
		printf("PE%d:\t", i);
		if (i == sender) {
			printf("p_bcastPutz(");
		} else {
			printf("p_bcastGetz(");
		}
		PRINT_ARRAY(8u, &(t[0]), len);
		printf(", ");
		print_32(len);
		printf(")\n");
	}
	putchar('\n');
}


void
bputgetz(void)
{
	register p_32 len = (p_randSame16u() % 15) + 1;
	register int source = (p_randUniq16u() % NPROC);
	register int source2;
	register int i, j;
	p_8u r[16], s[16], t[16], u[16];

	for (i=0; i<len; ++i) {
		r[i] = p_randUniq8u();
		s[i] = 0;
	}

	p_putGetz(&(s[0]), &(r[0]), len, source);

	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\tp_putGetz(", i);

		if (i == IPROC) {
			p_bcastPutz(&(s[0]), len);
			p_bcastPutz(&(r[0]), len);
			p_bcastPut32(source2 = source);
			for (j=0; j<len; ++j) {
				t[j] = s[j];
				u[j] = r[j];
			}
		} else {
			p_bcastGetz(&(t[0]), len);
			p_bcastGetz(&(u[0]), len);
			source2 = p_bcastGet32();
		}

		PRINT_ARRAY(8u, &(t[0]), len);
		printf(", ");
		PRINT_ARRAY(8u, &(u[0]), len);
		printf(", ");
		print_32(len);
		printf(", ");
		print_32(source2);
		printf(")\n");
	}
	putchar('\n');
}


void
bgatherz(void)
{
	register p_32 len = (p_randSame16u() % 4) + 1;
	register int i, j;
	p_8u r[5], s[5 * NPROC], t[5 * NPROC], u[5];

	for (i=0; i<len; ++i) {
		r[i] = p_randUniq8u();
		s[i] = 0;
	}

	p_gatherz(&(s[0]), &(r[0]), len);

	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\tp_gatherz(", i);

		if (i == IPROC) {
			p_bcastPutz(&(s[0]), (NPROC * len));
			p_bcastPutz(&(r[0]), len);
			for (j=0; j<(NPROC * len); ++j) {
				t[j] = s[j];
			}
			for (j=0; j<len; ++j) {
				u[j] = r[j];
			}
		} else {
			p_bcastGetz(&(t[0]), (NPROC * len));
			p_bcastGetz(&(u[0]), len);
		}

		PRINT_ARRAY(8u, &(t[0]), (NPROC * len));
		printf(", ");
		PRINT_ARRAY(8u, &(u[0]), len);
		printf(", ");
		print_32(len);
		printf(")\n");
	}
	putchar('\n');
}


void
bscatz(void)
{
	register p_32 len = (p_randSame16u() % 4) + 1;
	register int sender = (p_randSame16u() % NPROC);
	register int i, j;
	p_8u r[5], s[5 * NPROC], t[5 * NPROC], u[5];

	if (IPROC == sender) {
		for (i=0; i<(NPROC * len); ++i) {
			s[i] = p_randUniq8u();
		}
		p_scatPutz(&(r[0]), &(s[0]), len);
	} else {
		p_scatGetz(&(r[0]), len);
	}

	for (i=0; i<NPROC; ++i) {
		if (i == sender) {
			printf("PE%d:\tp_scatPutz(", i);
		} else {
			printf("PE%d:\tp_scatGetz(", i);
		}

		if (i == IPROC) {
			p_bcastPutz(&(r[0]), len);
			for (j=0; j<len; ++j) {
				u[j] = r[j];
			}
			p_bcastPutz(&(s[0]), (len * NPROC));
			for (j=0; j<len; ++j) {
				t[j] = s[j];
			}
		} else {
			p_bcastGetz(&(u[0]), len);
			p_bcastGetz(&(t[0]), (len * NPROC));
		}
		PRINT_ARRAY(8u, &(u[0]), len);

		if (i == sender) {
			printf(", ");
			PRINT_ARRAY(8u, &(t[0]), (len * NPROC));
		}
		printf(", %d)\n", len);
	}
	putchar('\n');
}


void
ball2allz(void)
{
	register p_32 len = (p_randSame16u() % 4) + 1;
	register int i, j;
	p_8u r[5 * NPROC], s[5 * NPROC], t[5 * NPROC], u[5 * NPROC];

	for (i=0; i<(len * NPROC); ++i) {
		r[i] = p_randUniq8u();
		s[i] = 0;
	}

	p_all2allz(&(s[0]), &(r[0]), len);

	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\tp_all2allz(", i);

		if (i == IPROC) {
			p_bcastPutz(&(s[0]), (NPROC * len));
			p_bcastPutz(&(r[0]), (NPROC * len));
			for (j=0; j<(NPROC * len); ++j) {
				t[j] = s[j];
				u[j] = r[j];
			}
		} else {
			p_bcastGetz(&(t[0]), (NPROC * len));
			p_bcastGetz(&(u[0]), (NPROC * len));
		}

		PRINT_ARRAY(8u, &(t[0]), (NPROC * len));
		printf(", ");
		PRINT_ARRAY(8u, &(u[0]), (NPROC * len));
		printf(", ");
		print_32(len);
		printf(")\n");
	}
	putchar('\n');
}


/*	Macro for functions of two arguments...
*/
#define	TWOARG(retsuf, name, argsuf1, argsuf2) \
{ \
	register p_##retsuf r = 0; \
	register p_##argsuf1 a1 = p_randUniq##argsuf1(); \
	register p_##argsuf2 a2 = p_randUniq##argsuf2(); \
	a2 = TWEAK(a2); \
	r = name##retsuf(a1, a2); \
	{ \
		register int i; \
		for (i=0; i<NPROC; ++i) { \
			printf("PE%d:\t", i); \
			if (i == IPROC) { \
				p_bcastPut##retsuf(r); \
				print_##retsuf(r); \
			} else { \
				print_##retsuf(p_bcastGet##retsuf()); \
			} \
			printf(" = " #name #retsuf "("); \
			if (i == IPROC) { \
				p_bcastPut##argsuf1(a1); \
				print_##argsuf1(a1); \
				printf(", "); \
				p_bcastPut##argsuf2(a2); \
				print_##argsuf2(a2); \
			} else { \
				print_##argsuf1(p_bcastGet##argsuf1()); \
				printf(", "); \
				print_##argsuf2(p_bcastGet##argsuf2()); \
			} \
			printf(")\n"); \
		} \
	} \
	putchar('\n'); \
}


/*	Macro for gather functions...
*/
#define	GATHER(suffix) \
{ \
	p_##suffix r[NPROC]; \
	register p_##suffix a = p_randUniq##suffix(); \
	p_gather##suffix(&(r[0]), a); \
	{ \
		p_##suffix t[NPROC]; \
		register int i; \
		for (i=0; i<NPROC; ++i) { \
			printf("PE%d:\t", i); \
			printf("p_gather" #suffix "("); \
			if (i == IPROC) { \
				p_bcastPutz(r, sizeof(t)); \
				PRINT_ARRAY(suffix, &(r[0]), NPROC); \
			} else { \
				p_bcastGetz(t, sizeof(t)); \
				PRINT_ARRAY(suffix, &(t[0]), NPROC); \
			} \
			printf(", "); \
			if (i == IPROC) { \
				p_bcastPut##suffix(a); \
				print_##suffix(a); \
			} else { \
				print_##suffix(p_bcastGet##suffix()); \
			} \
			printf(")\n"); \
		} \
	} \
	putchar('\n'); \
}


/*	Macro for scatter functions...
*/
#define	SCATTER(suffix) \
{ \
	p_##suffix a[NPROC]; \
	register p_##suffix r; \
	register int sender = (p_randSame16u() % NPROC); \
	register int i; \
	for (i=0; i<NPROC; ++i) a[i] = p_randUniq##suffix(); \
	if (sender == IPROC) { \
		r = p_scatPut##suffix(&(a[0])); \
	} else { \
		r = p_scatGet##suffix(); \
	} \
	{ \
		p_##suffix t[NPROC]; \
		for (i=0; i<NPROC; ++i) { \
			printf("PE%d:\t", i); \
			if (i == IPROC) { \
				p_bcastPut##suffix(r); \
				print_##suffix(r); \
			} else { \
				print_##suffix(p_bcastGet##suffix()); \
			} \
			if (i == sender) { \
				printf(" = p_scatPut" #suffix "("); \
				if (i == IPROC) { \
					p_bcastPutz(a, sizeof(t)); \
					PRINT_ARRAY(suffix, &(a[0]), NPROC); \
				} else { \
					p_bcastGetz(t, sizeof(t)); \
					PRINT_ARRAY(suffix, &(t[0]), NPROC); \
				} \
				printf(")\n"); \
			} else { \
				printf(" = p_scatGet" #suffix "()\n"); \
			} \
		} \
	} \
	putchar('\n'); \
}


/*	Macro for all2all functions...
*/
#define	ALL2ALL(suffix) \
{ \
	p_##suffix r[NPROC]; \
	p_##suffix s[NPROC]; \
	register int i; \
	for (i=0; i<NPROC; ++i) { \
		s[i] = p_randUniq##suffix(); \
	} \
	p_all2all##suffix(&(r[0]), &(s[0])); \
	{ \
		p_##suffix t[NPROC]; \
		register int i; \
		for (i=0; i<NPROC; ++i) { \
			printf("PE%d:\t", i); \
			printf("p_all2all" #suffix "("); \
			if (i == IPROC) { \
				p_bcastPutz(r, sizeof(t)); \
				PRINT_ARRAY(suffix, &(r[0]), NPROC); \
			} else { \
				p_bcastGetz(t, sizeof(t)); \
				PRINT_ARRAY(suffix, &(t[0]), NPROC); \
			} \
			printf(", "); \
			if (i == IPROC) { \
				p_bcastPutz(s, sizeof(t)); \
				PRINT_ARRAY(suffix, &(s[0]), NPROC); \
			} else { \
				p_bcastGetz(t, sizeof(t)); \
				PRINT_ARRAY(suffix, &(t[0]), NPROC); \
			} \
			printf(")\n"); \
		} \
	} \
	putchar('\n'); \
}


/*	Function to "tweak" argument
*/
#define	TWEAK(x)	x


void
control(void)
{
	register p_32 r;
	register int i;

	p_exit();
	r = p_init();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\tp_exit(); ", i);
		if (i == IPROC) {
			p_bcastPut32(r);
			print_32(r);
		} else {
			print_32(p_bcastGet32());
		}
		printf(" = p_init()\n");
	}
	putchar('\n');

	r = p_confirm();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (i == IPROC) {
			p_bcastPut32(r);
			print_32(r);
		} else {
			print_32(p_bcastGet32());
		}
		printf(" = p_confirm()\n");
	}
	putchar('\n');
}

void
primitives(void)
{
	register int i;

	p__wait();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\tp__wait()\n", i);
	}
	putchar('\n');

	p_wait();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\tp_wait()\n", i);
	}
	putchar('\n');

	ONEARG(bar, p__waitbar, 1u)
	ONEARG(bar, p_waitbar, 1u)
	ONEARG(1u, p__any, 1u)
	ONEARG(1u, p_any, 1u)
	ONEARG(1u, p__all, 1u)
	ONEARG(1u, p_all, 1u)
}


void
communications(void)
{
#define	P_DECLARE(suffix) \
PUTGET(suffix, p_bcast)
P_ALPHA
#undef	P_DECLARE

	bcastz();

#undef	TWEAK
#define	TWEAK(x)	(x % NPROC)
#define	P_DECLARE(suffix) \
TWOARG(suffix, p_putGet, suffix, 8u)
P_ALPHA
#undef	P_DECLARE
#undef	TWEAK
#define	TWEAK(x)	x

	bputgetz();

#define	P_DECLARE(suffix) \
GATHER(suffix)
P_ALPHA
#undef	P_DECLARE

	bgatherz();

#define	P_DECLARE(suffix) \
SCATTER(suffix)
P_ALPHA
#undef	P_DECLARE

	bscatz();

#define	P_DECLARE(suffix) \
ALL2ALL(suffix)
P_ALPHA
#undef	P_DECLARE

	ball2allz();
}


void
reductions(void)
{
#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceAnd##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceOr##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceXor##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceMin##suffix, suffix)
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceMax##suffix, suffix)
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceAdd##suffix, suffix)
P_GAMMA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_reduceMul##suffix, suffix)
P_GAMMA
#undef	P_DECLARE
}


void
scans(void)
{
#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanAnd##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanOr##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanXor##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanMin##suffix, suffix)
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanMax##suffix, suffix)
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanAdd##suffix, suffix)
P_GAMMA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(suffix, p_scanMul##suffix, suffix)
P_GAMMA
#undef	P_DECLARE
}


void
voting(void)
{
	ONEARG(32, p_first, 1u)
	ONEARG(32, p_count, 1u)
	ONEARG(32, p_quantify, 1u)

#undef	TWEAK
#define	TWEAK(x)	(x % (NPROC + 1))
	ONEARG(32, p_voteCount, 8u)
	ONEARG(bar, p_vote, 8u)
#undef	TWEAK
#define	TWEAK(x)	x

#define	P_DECLARE(suffix) \
ONEARG(32, p_matchCount##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(bar, p_match##suffix, suffix)
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
ONEARG(32, p_rank##suffix, suffix)
P_ALPHA
#undef	P_DECLARE
}


static p_bar dumm = 0;

void
dumsig(p_bar m)
{
	dumm |= m;
}

void
scontrol(void)
{
	register p_32 r;
	register int i;
	register p_32 dumsigno;
	register p_bar sigers = p_randSamebar();

	/* Start by installing a signal handler */
	s_wait();
	dumsigno = s_install(S_SIGANY, dumsig);
	s_wait();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (i == IPROC) {
			p_bcastPut32(dumsigno);
			print_32(dumsigno);
		} else {
			print_32(p_bcastGet32());
		}
		printf(" = s_install(S_SIGANY, dumsig)\n");
	}
	putchar('\n');
	if (s_init()) exit(1);

	/* Now invoke the signal handler */
	dumm = 0;
	s_wait();
	if (sigers & (((p_bar) 1) << (IPROC - 1))) {
		s__signal(dumsigno);
	}
	s_wait();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (sigers & (((p_bar) 1) << (i - 1))) {
			printf("s__signal(");
			if (i == IPROC) {
				p_bcastPut32(dumsigno);
				print_32(dumsigno);
			} else {
				print_32(p_bcastGet32());
			}
			printf("); ");
		}
		printf("dumsig(");
		if (i == IPROC) {
			p_bcastPutbar(dumm);
			print_bar(dumm);
		} else {
			print_bar(p_bcastGetbar());
		}
		printf(")\n");
	}
	putchar('\n');
	if (s_init()) exit(1);


	/* Now uninstall the signal handler */
	s_wait();
	dumsigno = s_install(dumsigno, s_ignore);
	s_wait();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (i == IPROC) {
			p_bcastPut32(dumsigno);
			print_32(dumsigno);
		} else {
			print_32(p_bcastGet32());
		}
		printf(" = s_install(%d, s_ignore)\n", dumsigno);
	}
	putchar('\n');
	if (s_init()) exit(1);

	/* Now invoke the signal handler */
	dumm = 0;
	s_wait();
	if (sigers & (((p_bar) 1) << (IPROC - 1))) {
		s_signal(dumsigno);
	}
	s_wait();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (sigers & (((p_bar) 1) << (i - 1))) {
			printf("s_signal(");
			if (i == IPROC) {
				p_bcastPut32(dumsigno);
				print_32(dumsigno);
			} else {
				print_32(p_bcastGet32());
			}
			printf("); ");
		}
		if (i == IPROC) {
			p_bcastPutbar(dumm);
			if (!dumm) printf("s_ignore()");
			else printf("dumsig() still called!");
		} else {
			if (!p_bcastGetbar()) printf("s_ignore()");
			else printf("dumsig() still called!");
		}
		printf("\n");
	}
	putchar('\n');
	if (s_init()) exit(1);


	s_exit();
	r = s_init();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\ts_exit(); ", i);
		if (i == IPROC) {
			p_bcastPut32(r);
			print_32(r);
		} else {
			print_32(p_bcastGet32());
		}
		printf(" = s_init()\n");
	}
	putchar('\n');
	if (s_init()) exit(1);
}


void
sprimitives(void)
{
	register int i;
	register p_1u j;
	register int dumsigno;

	/* The following magic tries to make the s__pending()
	   results consistent by skewing the timing of PEs
	*/
	dumsigno = s_install(S_SIGANY, dumsig);
	s_wait();
	s_wait();
	if (IPROC == (p_randSame16u() % NPROC)) {
		s_signal(dumsigno);
	} else {
		sleep(2);
	}
	j = s__pending();
	s_wait();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (IPROC == i) {
			p_bcastPut1u(j);
			print_1u(j);
		} else {
			print_1u(p_bcastGet1u());
		}
		printf(" = s__pending()\n");
	}
	putchar('\n');
	if (s_init()) exit(1);

	/* The following magic tries to make the s_pending()
	   results consistent by skewing the timing of PEs
	*/
	s_wait();
	s_wait();
	if (IPROC == (p_randSame16u() % NPROC)) {
		s_signal(dumsigno);
	} else {
		sleep(2);
	}
	j = s_pending();
	s_wait();
	s_exit();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\t", i);
		if (IPROC == i) {
			p_bcastPut1u(j);
			print_1u(j);
		} else {
			print_1u(p_bcastGet1u());
		}
		printf(" = s_pending()\n");
	}
	/* uninstall the "dum" signal handler */
	dumsigno = s_install(dumsigno, s_ignore);
	putchar('\n');
	if (s_init()) exit(1);

	s__poll();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\ts__poll()\n", i);
	}
	putchar('\n');

	s_poll();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\ts_poll()\n", i);
	}
	putchar('\n');

	s__wait();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\ts__wait()\n", i);
	}
	putchar('\n');

	s_wait();
	for (i=0; i<NPROC; ++i) {
		printf("PE%d:\ts_wait()\n", i);
	}
	putchar('\n');
}


int
main(register int argc,
register char **argv,
register char **arge)
{
	register p_64u seed = 0;
	register p_32 count = 1;

	if ((argc < 2) || (argc > 3)) {
		fprintf(stderr,
			"Usage: %s seed [repeat count]\n",
			argv[0]);
		exit(1);
	}

	/* check-in with AFAPI */
	if (p_init()) exit(1);

	if (IPROC == CPROC) {
		seed = atoi(argv[1]);
		p_bcastPut64u(seed);
		if (argc == 3) count = atoi(argv[2]);
		p_bcastPut32(count);
	} else {
		seed = p_bcastGet64u();
		count = p_bcastGet32();
	}

	/* init random generators */
	p_randInit(seed);

	/* run the tests */
	while (count > 0) {
		/* Test all the p_ stuff */
		control();
		primitives();
		communications();
		reductions();
		scans();
		voting();

		/* Test the basic s_ stuff */
		if (s_init()) exit(1);
		scontrol();
		sprimitives();
		s_exit();

		--count;
	}

	/* Exit cleanly */
	p_exit();
	exit(0);
}
