/*	config.h

	AFAPI compile-time configuration information.
	There are four sections below:

	1) SHMAPERS specific configurations
	2) Non-SHMAPERS configurations
	3) Universal configurations
	4) Video Wall configurations

	July 1997 by H. Dietz and T.I. Mattox
*/


#ifndef _P_CONFIG_H
#define _P_CONFIG_H

/*	1) SHMAPERS specific configurations

	It is important to note that, although SHMAPERS library still
	refers to PEs/processors, each is really just a UNIX process.
	Thus, under SMP OS like Linux, each process might not actually
	stay with a particular processor.  In fact, you can even lie
	about how many PEs/processors you have...  although the OS
	scheduling is likely to make your programs much more efficient
	if you do not lie....

	define P_SHM_IDLE {}
		If present, this sets the shared memory system to use {} as
		the OS idle code.  By default, a usleep(1) is inserted 
		if the NPROC and SHMPROC constants indicate that there are
		more PEs than CPUs to lessen potential spin lock time waiting
		for a process that the OS isn't running.

		Generally, NPROC should equal the ACTUAL NUMBER of processors
		in your system...  but heavily loaded systems will do better
		with smaller values and larger values can be useful
		for testing....
*/
#define	P_SHM_IDLE {}
#undef	P_SHM_IDLE


/*	2) Non-SHMAPERS configurations

	Software debounce option...

	Leave P_DEBOUNCE defined if your port hardware otherwise has
	noise-related errors due to God-only-knows-what.  The solution
	used is to make each port input operation keep sampling until
	the same value has been read twice in a row, which should be
	plenty of settling time since this noise generally settles in
	less than 1/10 of the time it takes for a port access.  You
	really should not need P_DEBOUNCE defined.
*/
#define	P_DEBOUNCE
#undef	P_DEBOUNCE

/*	I/O permisions option: iopl() vs. ioperm()

	Leave P_USE_IOPL undefined if your port hardware is in the normal
	range of printer port addresses.  If you have an add-on PCI parallel
	port card, you may need to define P_USE_IOPL to gain access to
	the port.
*/
#define	P_USE_IOPL
#undef	P_USE_IOPL

/*	Non-Standard Port address option: 

	Leave P_NON_STANDARD_PORT undefined if your port hardware is at one
	of the three normal of printer port addresses: 0x278, 0x378, 0x3bc
	If you have an add-on parallel port card, you may need to
	define P_NON_STANDARD_PORT to gain access to the port.

	If you have P_USE_IOPL set, this will be activated as well.
*/
#ifdef	P_USE_IOPL
#define	P_NON_STANDARD_PORT
#else
#define	P_NON_STANDARD_PORT
#undef	P_NON_STANDARD_PORT
#endif


/*
	3) Universal configurations

	The magic "constants"...
*/
#define	P_CACHELINE 32	/* Number of bytes in a cache line */

#ifndef	NPROC		/* Should be defined by the Makefile */
#define	NPROC	4	/* Number of PEs */
#endif
#ifndef	SMPPROC		/* Should be defined by the Makefile */
#define	SMPPROC	1	/* Number of actual CPUs available for SHMAPERS */
#endif

/*	4) Video Wall configurations

	The following two settings determine the maximum size screen
	that the VWLIB will try to obtain from the underlying graphics
	library (X, Frame Buffer, SVGALib).

	WARNING:
	Setting them above the default of 640x480 may cause the VGA
	library to request a screen resolution beyond the ratings
	of your monitor, BE CAREFUL!

	VWWIDE	Maximum width (in pixels) of any screen
	VWTALL	Maximum height (in pixels) of any screen
*/
#define VWWIDE  640    /* maximum width, of any screen */
#define VWTALL  480    /* maximum height, of any screen */


/*	VWMAXWALLS is the maximum number of walls allowed in the Space
*/
#define	VWMAXWALLS	((NPROC * 2) + 1)

/*	VWCONVERT is the fully specified pathname for ImageMagik 'convert'

	The 'convert' program is used to generate P6 .ppm files
	from other image formats.  It is assumed that it will not
	generate P3 files, but this may not always be true.
	To be safe, always feed the VWLIB raw P6 .ppm files directly.
*/
#define	VWCONVERT	"/usr/bin/convert"

/*	Mouse Configuration:  These currently are not used for X based VWLIB


	VW_MOUSE_DEV	The mouse device pathname
	VW_MOUSE_TYPE	The mouse device type, usually one of:
			MOUSE_MICROSOFT
			MOUSE_IMPS2
	See /etc/vga/libvga.config or `man vga_getmousetype` for a full list

	VW_MOUSE_XRES	The range of distinct mouse positions in horiz. dim.
	VW_MOUSE_YRES	The range of distinct mouse positions in vert. dim.
	VW_MOUSE_WRAP	The mode for wrapping at the edges, usually one of:
			MOUSE_WRAPX
			MOUSE_WRAPY
			MOUSE_NOWRAP
*/
#define	VW_MOUSE_DEV	"/dev/mouse"
#define	VW_MOUSE_TYPE	MOUSE_IMPS2
#define	VW_MOUSE_XRES	51200
#define	VW_MOUSE_YRES	51200
#define	VW_MOUSE_WRAP	MOUSE_WRAPX


/*	VWVGA_BLIT configuration

	If VWVGA_BLIT is defined, the VGA version of the VWLib will
	update the video memory in linear order, as fast as possible.
	However, this can result in temporary visual artifacts when the
	scan line of your monitor intersects with where the blit is
	occuring, and is especially noticable on horizontally moving
	images with sharp vertical features.

	If you want to avoid this motion artifact, put #undef VWVGA_BLIT
	after the #define below to make the VWLib use a deliberately
	random order to paint pixels; this minimizes the effects of the
	race between display retrace and repaint.  Blits are faster,
        but have nasty artifacts.
*/
#define VWVGA_BLIT


/*	VWTURN_OFF_CURSOR configuration for Frame Buffer implementation

	If VWTURN_OFF_CURSOR is defined, the setterm unix command
	will be used to turn off the cursor while the Frame Buffer VWLib
	is being used.  It will restore the cursor in vwend().  However,
	if your program does not exit properlu, you will be left without
	a text cursor.  To avoid this problem, put #undef VWTURN_OFF_CURSOR
	after the #define to leave the cursor alone.

*/
#define	VWTURN_OFF_CURSOR


#endif
