/*	vwsup.h

	Video Wall SUPport header file.

	Termnology:

	SPACE	The set of all walls

	WALL	A set of individual PE display panels,
		treated as a single logical entity

	PANEL	An individual PE display

	I.e., a SPACE is defined by a set of WALLs and
	a WALL is defined by a set of PANELs

	February 1999 by H. Dietz
*/


#include <AFAPI/afapi.h>


typedef	unsigned char	pixel8;
typedef	unsigned short	pixel16;
typedef	unsigned char	pixel24;
typedef	unsigned int	pixel32;

typedef struct {
	int	x, y;
} vwxy;

typedef struct {
	vwxy	min;
	vwxy	max;
} vwbox;

typedef struct {
	float	x, y;
} vwfxy;

typedef struct {
	vwfxy	min;
	vwfxy	max;
} vwfbox;

typedef struct {
	vwxy	dim;		/* buffer dimensions */
	vwbox	dirty;		/* dirty region (clean dirty.min.x < 0) */
	int	depth;		/* pixel depth: 8, 16, 24, 32 */
	int	bytes;		/* data size in bytes */
	union {			/* pointer to actual data */
		pixel8	*data8;
		pixel16	*data16;
		pixel24	*data24;
		pixel32	*data32;
		void	*data;
	} data;
} vwbuf;

vwbuf	vwpanel;		/* this PE's panel buffer */

typedef struct {
	char	*name;		/* name of this wall */
	vwfbox	all;		/* dimensions of the wall */
	vwfbox	one[NPROC];	/* position of each panel */
} vwwall;

vwwall	*vwspace[NPROC];	/* which wall each PE belongs to */


/*	Video Wall Main stuff...
*/

extern void
vwinit(void);

extern void
vwdefWalls(register char *filename);

extern vwwall
*vwgetWall(register char *name);

extern void
vwdefSpace(register vwwall *wall);

extern void
vwexit(void);

extern void
vpclear(register vwbuf *panel);

extern void
vwclear(register vwwall *wall);


/*	Video Wall Display Device routines...
*/

extern void
vwbegin(void);

extern void
vppaint(register vwbuf *panel);

extern void
vwpaint(register vwwall *wall);

extern void
vwend(void);


/*	Video Wall RESample routines...
*/

extern void
vpres(register vwbuf *destpanel,
register vwbox destarea,
register vwbuf *srcimage,
register vwfbox srcarea,
register int frac);

extern void
vwres(register vwwall *destwall,
register vwfbox destarea,
register vwbuf *srcimage,
register vwfbox srcarea,
register int frac);



/*	Video Wall MAGnification routines...
*/

extern void
vwscale(register vwwall *destwall,
register vwfbox destarea,
register vwfxy destctr,
register vwbuf *srcimage,
register vwfxy srcctr,
register vwfxy srcmag,
register int frac);

extern float
vpaspect(register vwbuf *panel);

extern float
vwaspect(register vwwall *destwall,
register vwfbox destarea);

extern void
vwmag(register vwwall *destwall,
register vwfbox destarea,
register vwfxy destctr,
register vwbuf *srcimage,
register vwfxy srcctr,
register float srcmag,
register int frac);


/*	Video Wall file stuff...
	currently supports only 24-bit P6 format PPMs
*/

extern vwbuf
vwmapFile(register char *filename);

extern vwbuf
vpmapFile(register char *filename);

extern void
vwunmapFile(register char *filename);

extern void
vpunmapFile(register char *filename);

extern int
vwwriteFile(register char *filename,
vwbuf image);


/*	Video Wall timer stuff...
	all times are measured in float seconds since vwreftime().
*/

typedef double vwtime;

extern vwtime
vwnow(void);

extern void
vwreftime(void);

#define	VWONTIME	0
#define	VWLATE		1

extern int
vwwait(register vwtime deadline);


/*	Video Wall MOUSE stuff...
*/

typedef struct {
	vwfxy	pos;		/* x,y position, 0:1,0:1 */
	int	left;		/* left button */
	int	center;		/* center button */
	int	right;		/* right button */
} vwinfoMouse;

extern void
vpbeginMouse(void);

extern vwinfoMouse
vpreadMouse(register vwtime until);

extern void
vpendMouse(void);

extern void
vwbeginMouse(void);

extern vwinfoMouse
vwreadMouse(register vwtime until);

extern void
vwendMouse(void);


/*	Video Wall KEYBoaRD stuff...
*/

#define	VWEOF		-1	/* the EOF marker was read */
#define	VWNOKEY		-2	/* no key was pressed */
#define	VWNOKEYBRD	-3	/* no keyboard is open */

extern void
vwbeginKeybrd(void);

extern int
vwreadKeybrd(register vwtime until);

extern void
vwendKeybrd(void);
