/*	one.c

	Complete "dummy" AFAPI for 1 process(or).

	December 1999 by Hank Dietz
*/

#include <AFAPI/afapi.h>

#include <stdio.h>
#include <string.h>
#include "one.h"

int iproc = 0;	/* PE number (0..NPROC-1) */
int cproc = 0;	/* Console PE number */

inline void
p_wait(void)
{
	/* Nobody else to wait for */
}

inline p_bar
p_waitbar(register p_1u f)
{
	return((p_bar) (f != 0));
}

inline p_1u
p_any(register p_1u f)
{
	return((p_1u) (f != 0));
}

inline p_1u
p_all(register p_1u f)
{
	return((p_1u) (f != 0));
}

inline int
p_init(void)
{
	/* Don't even look at the environment...
	   just say it worked.
	*/
	return(0);
}

inline void
p_exit(void)
{
	/* Nothing to do... */
}

inline int
p_confirm(void)
{
	/* Everything is healthy... */
	return(0);
}

/*	p_bcastPut/Get() for all data types...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_bcastPut##suffix(register p_##suffix d) \
{ \
	return(d); \
} \
 \
inline p_##suffix \
p_bcastGet##suffix(void) \
{ \
		fprintf(stderr, \
"AFAPI PE:  cannot p_bcastGet with only 1 PE\n"); \
	return((p_##suffix) 0); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_bcastPutz() and p_bcastGetz()...
*/

inline void
p_bcastPutz(register void *source,
register int bytes)
{
	/* If a message is sent and there is nobody there
	   to see it, we don't need to do anything...
	*/
}


inline void
p_bcastGetz(register void *dest,
register int bytes)
{
		fprintf(stderr,
"AFAPI PE:  cannot p_bcastGetz with only 1 PE\n");
}

/*	p_putGet() functions...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_putGet##suffix(register p_##suffix datum, \
register int source) \
{ \
	return(datum); \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_putGetz(register void *dest,
register void *src,
register int n,
register int source)
{
	if (dest) {
		while (n-- > 0) {
			*((p_8u *) dest++) = p_putGet8u(*((p_8u *) src++), source);
		}
	}
}

/*	p_gather() functions...
*/

#define	P_DECLARE(suffix) \
inline void \
p_gather##suffix(register p_##suffix *dest, \
register p_##suffix datum) \
{ \
	*dest = datum; \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_gatherz(register void *dest,
register void *src,
register int n)
{
	if (dest) {
		memcpy(dest, src, n);
	}
}

/*	p_scatPut() and p_scatGet() functions...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scatPut##suffix(register p_##suffix *src) \
{ \
	return(*src); \
} \
 \
inline p_##suffix \
p_scatGet##suffix(void) \
{ \
		fprintf(stderr, \
"AFAPI PE:  cannot p_scatGet with only 1 PE\n"); \
	return((p_##suffix) 0); \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_scatPutz(register void *dest,
register void *src,
register int n)
{
	if (dest) {
		memcpy(dest, src, n);
	}
}

inline void
p_scatGetz(register void *dest,
register int n)
{
		fprintf(stderr, \
"AFAPI PE:  cannot p_scatGetz with only 1 PE\n"); \
}

/*	p_all2all() functions...
*/
#define	P_DECLARE(suffix) \
inline void \
p_all2all##suffix(register p_##suffix *dest, \
register p_##suffix *src) \
{ \
	if (dest) { \
		*dest = *src; \
	} \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_all2allz(register void *dest,
register void *src,
register int n)
{
	register int dist = 1;

	if (dest) {
		memcpy(dest, src, n);
	}
}

/*	p_reduceAnd() functions...
*/

inline p_1u
p_reduceAnd1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceAnd##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)

#undef	P_DECLARE

/*	p_reduceOr() functions...
*/

inline p_1u
p_reduceOr1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceOr##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)

#undef	P_DECLARE

/*	p_reduceXor() functions...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceXor##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_reduceMin() functions...
*/

inline p_1u
p_reduceMin1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceMax() functions...
*/

inline p_1u
p_reduceMax1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceAdd() functions...
*/

inline p_1u
p_reduceAdd1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceAdd##suffix(register p_##suffix datum) \
{ \
	return(P_CAST_##suffix(datum)); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceMul() functions...
*/

inline p_1u
p_reduceMul1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMul##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanAnd() functions...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanAnd##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_scanOr() functions...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanOr##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_scanXor() functions...
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanXor##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_scanMin() functions...
*/

inline p_1u
p_scanMin1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanMin##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanMax() functions...
*/

inline p_1u
p_scanMax1u(register p_1u datum)
{
	return(p_scanOr1u(datum));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanMax##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanAdd() functions...
*/

inline p_1u
p_scanAdd1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanAdd##suffix(register p_##suffix datum) \
{ \
	return(P_CAST_##suffix(datum)); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)


#undef P_DECLARE

/*	p_scanMul() functions...
*/

inline p_1u
p_scanMul1u(register p_1u datum)
{
	return((p_1u) (datum != 0));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_scanMul##suffix(register p_##suffix datum) \
{ \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef P_DECLARE

/*	p_first()...
*/

inline int
p_first(register p_1u f)
{
	return(f ? 0 : NPROC);
}

/*	p_count()...

	Count how many voted yes.
*/

inline int
p_count(register p_1u f)
{
	return(f ? 1 : 0);
}

/*	p_quantify()...
*/

inline int
p_quantify(register p_1u f)
{
	return(f ? NPROC : 0);
}

/*	p_vote()...
*/

inline p_bar
p_vote(register int pe)
{
	return((p_bar) (pe == IPROC));
}

/*	p_voteCount()...
*/

inline int
p_voteCount(register int pe)
{
	return(pe == IPROC);
}

/*	p_match()...
*/

#define	P_DECLARE(suffix) \
inline p_bar \
p_match##suffix(register p_##suffix datum) \
{ \
	return((p_bar) 1); \
}

P_BETA

#undef	P_DECLARE

/*	p_matchCount()...
*/

#define	P_DECLARE(suffix) \
inline int \
p_matchCount##suffix(register p_##suffix datum) \
{ \
	return(1); \
}

P_BETA

#undef	P_DECLARE

/*	p_rank()...
*/

#define	P_DECLARE(suffix) \
inline int \
p_rank##suffix(register p_##suffix datum) \
{ \
	return(0); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_ranked()...
*/

#define	P_DECLARE(suffix) \
inline int \
p_ranked##suffix(register p_##suffix datum, \
register int rank) \
{ \
	return(0); \
}

P_ALPHA

#undef	P_DECLARE


#define	S_SIGMAX	5
#define	S_EXIT		(S_SIGMAX - 1)


static int s_active = 0;	/* Signals enabled? */

/*	Pre-defined signal handlers...
*/

void
s_ignore(register p_bar m)
{
	/* Nothing to do...  that's why it's called s_ignore() */
}


static void
s_exitHandler(register p_bar m)
{
	/* The one-and-only guy wants out; do it */
	s_active = 0;
}

/*	s_vector[] array...

	The following is the magic internal structure used to
	vector signals to the correct handlers.  The entry for
	s_exitHandler must be in the S_SIGMAX-1 position.
*/

static void (*s_vector[S_SIGMAX])(register p_bar mask) = {
	/* Normal signals... */
	s_ignore,
	s_ignore,
	s_ignore,
	s_ignore,

	/* And one signal you can't redefine... */
	s_exitHandler
};

/*	s_install()
*/

int
s_install(register int s,
register void (*fn)(p_bar m))
{
	/* Install the signal handler fn()... */
	register int which = s;

	/* If s is S_SIGANY, pick one that is s_ignore()... */
	if (which == S_SIGANY) {
		for (which = 0; which<(S_SIGMAX-1); ++which) {
			if (s_vector[which] == s_ignore) break;
		}
	}

	/* Got a valid signal number? */
	if ((which < 0) || (which >= (S_SIGMAX-1))) {
		/* Nope.  Return with error. */
		return(-1);
	}

	/* Ok, install it... */
	s_vector[which] = fn;

	/* Return signal number */
	return(which);
}

/*	s_init() and s_exit()...
*/

int
s_init(void)
{
	/* If it's already active, oops... */
	if (s_active) return(1);

	/* Ok, make it active... */
	s_active = 1;
	return(0);
}


void
s_exit(void)
{
	/* Shut-down the signal system using a signal... */
	while (s_active) {
		s_signal(S_EXIT);
	}

	/* Then make sure that everybody got out... */
	p_wait();
}

/*	s_signal()...

	This code must recognize who has sent signals, create masks grouping
	similar signals together, and invoke the appropriate handler
	functions in a way that gives s_exit the lowest priority....
*/

void
s_signal(register int s)
{
	p_32 sigs[NPROC];
	register int sig;

	/* Send our signal and collect results... */
	p_gather32(&(sigs[0]), s);

	/* Scan for signals in priority order... */
	for (sig=0; sig<S_SIGMAX; ++sig) {
		register p_bar mask = 0;
		register p_bar bit = 1;
		register int i;

		/* Collect similar signals... */
		for (i=0; i<NPROC; ++i) {
			if (sigs[i] == sig) {
				mask |= bit;
			}

			bit <<= 1;
		}

		/* If we found somebody, call the handler... */
		if (mask != 0) {
			s_vector[sig](mask);
		}
	}
}

p_1u
s_pending(void)
{
	/* Nobody is here yet... i.e., we're not */
	return(0);
}

inline void
s_poll(void)
{
	/* Nothing to do */
}


inline void
s_wait(void)
{
	/* Nothing to do */
}
