/*	reduce.c

	AFAPI associative reduction function definitions for
	SHMAPERS.

	July 1996 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"

/*	p_reduceAnd() functions...
*/

inline p_1u
p_reduceAnd1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceAnd##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		((p_##suffix##u) datum) &= \
			((p_##suffix##u) p->datum[ which ].u_##suffix); \
	} while (++p < q); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)

#undef	P_DECLARE

/*	p_reduceOr() functions...
*/

inline p_1u
p_reduceOr1u(register p_1u datum)
{
	return(p__any(datum));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceOr##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		((p_##suffix##u) datum) |= \
			((p_##suffix##u) p->datum[ which ].u_##suffix); \
	} while (++p < q); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)

#undef	P_DECLARE

/*	p_reduceXor() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceXor##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	datum = 0; \
	p = p_shm_all; \
	do { \
		((p_##suffix##u) datum) ^= \
			((p_##suffix##u) p->datum[ which ].u_##suffix); \
	} while (++p < q); \
 \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_reduceMin() functions...
*/

inline p_1u
p_reduceMin1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		register p_##suffix t = p->datum[ which ].u_##suffix; \
		if (t < datum) datum = t; \
	} while (++p < q); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceMax() functions...

	This is essentially the same as the p_reduceMin() stuff.
*/

inline p_1u
p_reduceMax1u(register p_1u datum)
{
	return(p__any(datum));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		register p_##suffix t = p->datum[ which ].u_##suffix; \
		if (t > datum) datum = t; \
	} while (++p < q); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceAdd() functions...
*/

inline p_1u
p_reduceAdd1u(register p_1u datum)
{
	return(p__any(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceAdd##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	datum = 0; \
	do { \
		datum += p->datum[ which ].u_##suffix; \
	} while (++p < q); \
 \
	return(P_CAST_##suffix(datum)); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceMul() functions...
*/

inline p_1u
p_reduceMul1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMul##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	datum = 1; \
	do { \
		datum *= p->datum[ which ].u_##suffix; \
	} while (++p < q); \
 \
	return(datum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

#endif /* if NPROC > 1 */
