/*	signal.c

	AFAPI signal definitions for SHMAPERS.

	July 1996 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"


#define	S_SIGMAX	5
#define	S_EXIT		(S_SIGMAX - 1)


static int s_active = 0;	/* Signals enabled? */

/*	Pre-defined signal handlers...
*/

void
s_ignore(register p_bar m)
{
	/* Nothing to do...  that's why it's called s_ignore() */
}


static void
s_exitHandler(register p_bar m)
{
	/* Is every process making this request? */
	if (m == ((((p_bar) 1) << NPROC) - ((p_bar)1))) {
		/* Yup, everybody wants out...  do it */
		s_active = 0;
	}
}

/*	s_vector[] array...

	The following is the magic internal structure used to vector
	signals to the correct handlers.  It is specific to SHMAPERS.
	The entry for s_exitHandler must be in the S_SIGMAX-1 position.
*/

static void (*s_vector[S_SIGMAX])(register p_bar mask) = {
	/* Normal signals... */
	s_ignore,
	s_ignore,
	s_ignore,
	s_ignore,

	/* And one signal you can't redefine... */
	s_exitHandler
};

/*	s_install()

	Install a signal handler...
*/

int
s_install(register int s,
register void (*fn)(p_bar m))
{
	/* Install the signal handler fn()... */
	register int which = s;

	/* If s is S_SIGANY, pick one that is s_ignore()... */
	if (which == S_SIGANY) {
		for (which = 0; which<(S_SIGMAX-1); ++which) {
			if (s_vector[which] == s_ignore) break;
		}
	}

	/* Got a valid signal number? */
	if ((which < 0) || (which >= (S_SIGMAX-1))) {
		/* Nope.  Return with error. */
		return(-1);
	}

	/* Ok, install it... */
	s_vector[which] = fn;

	/* Return signal number */
	return(which);
}

/*	s_init() and s_exit()...
*/

int
s_init(void)
{
	/* If it's already active, oops... */
	if (s_active) return(1);

	/* Ok, make it active... */
	s_active = 1;
	return(0);
}


void
s_exit(void)
{
	/* Shut-down the signal system using a signal... */
	while (s_active) {
		s_signal(S_EXIT);
	}

	/* Then make sure that everybody got out... */
	p_wait();
}

/*	s_signal()...

	This code must recognize who has sent signals, create masks grouping
	similar signals together, and invoke the appropriate handler
	functions in a way that gives s_exit the lowest priority....
*/

void
s_signal(register int s)
{
	p_32 sigs[NPROC];
	register int sig;

	/* Send our signal and collect results... */
	p_gather32(&(sigs[0]), s);

	/* Scan for signals in priority order... */
	for (sig=0; sig<S_SIGMAX; ++sig) {
		register p_bar mask = 0;
		register p_bar bit = 1;
		register int i;

		/* Collect similar signals... */
		for (i=0; i<NPROC; ++i) {
			if (sigs[i] == sig) {
				mask |= bit;
			}

			bit <<= 1;
		}

		/* If we found somebody, call the handler... */
		if (mask != 0) {
			s_vector[sig](mask);
		}
	}
}

/*	Inlined signal stuff...

	These are just callable versions of the simple interface
	routines....
*/

p_1u
s_pending(void)
{
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = (p_shm_me->barno + 1);
	register int which = (barno & 1);

	/* Is anyone at the next barrier, and not sending S_EXIT? */
	do {
		/* If so, return 1... */
		if ((p->barno == barno) && (p->datum[which].u_32 != S_EXIT)){
			return(1);
		}
	} while (++p < q);

	/* Nobody is here yet */
	return(0);
}


void
s_poll(void)
{
	s__poll();
}


void
s_wait(void)
{
	s__wait();
}
#endif /* if NPROC > 1 */
