/*	vote.c

	AFAPI voting and scheduling function definitions for
	SHMAPERS.

	July 1996 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"

/*	p_first()...

	Return the IPROC of the first lowest-numbered PE with a yes
	vote.  For SHMAPERS, this is very similar to p_any().
*/

int
p_first(register p_1u f)
{
	/* Collect the array of flags and look for a 1... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for a 1... */
	p = p_shm_all;
	do {
		if (p->datum[ which ].u_1u) return(p - p_shm_all);
	} while (++p < q);

	return(NPROC);
}

/*	p_count()...

	Count how many voted yes.
*/

inline int
p_count(register p_1u f)
{
	/* Collect the array of flags and look for a 1... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register int count = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for 1 values and count them... */
	p = p_shm_all;
	do {
		count += p->datum[ which ].u_1u;
	} while (++p < q);

	return(count);
}

/*	p_quantify()...

	Quantify the number of PEs voting yes; the return is 0, 1, 2
	(really >1), or NPROC.  If NPROC is 3 or less, p_quantify
	returns an exact count.  As of the 9807xx version of AFAPI,
	p_quantify() will return identical values regardless of which
	implementation of AFAPI is used.
*/

int
p_quantify(register p_1u f)
{
	register int ct = p_count(f);
	if ((ct > 2) && (ct < NPROC)) {
		return(2);
	} else {
		return(ct);
	}
}

/*	p_vote()...

	Builds a vote barrier bit vector much as p_waitbar().
*/

inline p_bar
p_vote(register int pe)
{
	/* Collect the array of votes and construct bit vector... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register p_bar result = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_32 = pe;
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	p = p_shm_all;
	do {
		--q;
		result = ((result << 1) |
			  (q->datum[ which ].u_32 == IPROC));
	} while (q > p);

	return(result);
}

/*	p_voteCount()...
*/

int
p_voteCount(register int pe)
{
	/* Collect the array of votes and count matches... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register int result = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_32 = pe;
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	p = p_shm_all;
	do {
		--q;
		result += (q->datum[ which ].u_32 == IPROC);
	} while (q > p);

	return(result);
}

/*	p_match()...
*/

#define	P_DECLARE(suffix) \
inline p_bar \
p_match##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
	register p_bar result = 0; \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Construct bit vector... */ \
	p = p_shm_all; \
	do { \
		--q; \
		result = ((result << 1) | \
			  (q->datum[ which ].u_##suffix == datum)); \
	} while (q > p); \
 \
	return(result); \
}

P_BETA

#undef	P_DECLARE

/*	p_matchCount()...
*/

#define	P_DECLARE(suffix) \
int \
p_matchCount##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
	register int result = 0; \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Count matches... */ \
	p = p_shm_all; \
	do { \
		--q; \
		result += (q->datum[ which ].u_##suffix == datum); \
	} while (q > p); \
 \
	return(result); \
}

P_BETA

#undef	P_DECLARE

/*	p_rank()...
*/

#define	P_DECLARE(suffix) \
int \
p_rank##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
	register int result = 0; \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Count things ranking below us... */ \
	p = p_shm_all; \
	do { \
		register p_##suffix t = p->datum[ which ].u_##suffix; \
		result += ((t < datum) || \
			   ((t == datum) && (p < p_shm_me))); \
	} while (++p < q); \
 \
	return(result); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_ranked()...

	Well, rank was such a handy operation on the MasPar...  it
	never occurred to us that it is backward.  Basically, if you
	want to sort using putGet, you don't care where your data
	goes, but simply where the datum you should get will have to
	come from.  The ranked operation generalizes this so that each
	PE can get the current location of the value that would be
	ranked as the second argument specifies.  This is currently
	done by a dumb gather and (rank-truncated) bubble sort.
*/

#define	P_DECLARE(suffix) \
int \
p_ranked##suffix(register p_##suffix datum, \
register int rank) \
{ \
	p_##suffix buf[NPROC]; \
	int index[NPROC]; \
	register int i, j; \
 \
	p_gather##suffix(&(buf[0]), datum); \
	if (rank >= NPROC) return(NPROC); \
	for (i=0; i<NPROC; ++i) index[i] = i; \
	for (i=0; i<=rank; ++i) { \
		for (j=i+1; j<NPROC; ++j) { \
			if (buf[index[i]] > buf[index[j]]) { \
				register int t = index[i]; \
				index[i] = index[j]; \
				index[j] = t; \
			} \
		} \
	} \
 \
	return(index[rank]); \
}

P_ALPHA

#undef	P_DECLARE
#endif /* if NPROC > 1 */
