/*	com.c

	AFAPI communication function definitions for TTL_PAPERS.

	July 1996 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"

/*	p_bcastPut/Get() for 1u, 8u, 8, and bar...
*/

inline p_1u
p_bcastPut1u(register p_1u datum)
{
	datum = P_CAST_1u(datum);
	p__put(datum);
	return(datum);
}


inline p_1u
p_bcastGet1u(void)
{
	return(p__get());
}


inline p_8u
p_bcastPut8u(register p_8u datum)
{
	p__put(datum >> 4);
	p__put(datum);
	return(datum);
}

inline p_8u
p_bcastGet8u(void)
{
	register p_8u high = p__get();
	return(p__get() | (high << 4));
}

inline p_8
p_bcastPut8(register p_8 datum)
{
	p__put(datum >> 4);
	p__put(datum);
	return(datum);
}

inline p_8
p_bcastGet8(void)
{
	register p_8 high = p__get();
	return(p__get() | (high << 4));
}

p_bar
p_bcastPutbar(register p_bar datum)
{
	register int bits = ((NPROC - 1) & ~0x3);

	do {
		/* p__put() masks to low 4 bits */
		p__put(datum >> bits);
	} while ((bits -= 4) >= 0);
	return(datum);
}

p_bar
p_bcastGetbar(void)
{
	register int bits = ((NPROC - 1) & ~0x3);
	register p_bar result = 0;

	do {
		result |= (((p_bar) p__get()) << bits);
	} while ((bits -= 4) >= 0);

	return(result);
}

/*	p_bcastPut/Get() for 16u, 16, 32, 32u, 64, 64u...

	Each of these routines uses a nybble-prefix tag to efficiently
	take advantage of the fact that the most significant bits of
	most 16 bit or larger integers are very often all 0s or all 1s.
	We can even increase the odds of this by XOR with the previous
	value (helpfull mostly for signed values and pointers).
	Imagine dividing each integer into four segments.  The first
	tag bit indicates if the most significant quarter is 0.  The
	remaining three bits indicate if the previous quarter and the
	next quarter have the same value.  64-bit stuff is actually
	treated as two 32-bit values.
*/
	
p_16u
p_bcastPut16u(register p_16u datum)
{
	/* Use simple compression algorithm...
	   4-bit tag says which nybbles differ
	*/
	register p_8u c0 = (datum & 0xf);
	register p_8u c1 = ((datum >> 4) & 0xf);
	register p_8u c2 = ((datum >> 8) & 0xf);
	register p_8u c3 = ((datum >> 12) & 0xf);
	register int tag = (((c0 == c1) ? 0 : 1) |
			    ((c1 == c2) ? 0 : 2) |
			    ((c2 == c3) ? 0 : 4) |
			    ((c3 == 0) ? 0 : 8));
	p__put(tag);
	if (tag & 8) p__put(c3);
	if (tag & 4) p__put(c2);
	if (tag & 2) p__put(c1);
	if (tag & 1) p__put(c0);

	return(datum);
}

p_16u
p_bcastGet16u(void)
{
	/* Use simple compression algorithm...
	   4-bit tag says which nybbles differ
	*/
	register p_16u datum = 0;
	register p_8u c;
	register int tag = p__get();

	datum = (c = ((tag & 8) ? p__get() : 0));
	if (tag & 4) c = p__get();
	datum = ((datum << 4) | c);
	if (tag & 2) c = p__get();
	datum = ((datum << 4) | c);
	if (tag & 1) c = p__get();
	return((datum << 4) | c);
}


p_32u
p_bcastPut32u(register p_32u datum)
{
	/* Use simple compression algorithm...
	   4-bit tag says which bytes differ
	*/
	register p_8u c0 = (datum & 0xff);
	register p_8u c1 = ((datum >> 8) & 0xff);
	register p_8u c2 = ((datum >> 16) & 0xff);
	register p_8u c3 = ((datum >> 24) & 0xff);
	register int tag = (((c0 == c1) ? 0 : 1) |
			    ((c1 == c2) ? 0 : 2) |
			    ((c2 == c3) ? 0 : 4) |
			    ((c3 == 0) ? 0 : 8));
	p__put(tag);
	if (tag & 8) p_bcastPut8u(c3);
	if (tag & 4) p_bcastPut8u(c2);
	if (tag & 2) p_bcastPut8u(c1);
	if (tag & 1) p_bcastPut8u(c0);

	return(datum);
}


p_32u
p_bcastGet32u(void)
{
	/* Use simple compression algorithm...
	   4-bit tag says which bytes differ
	*/
	register p_32u datum = 0;
	register p_8u c;
	register int tag = p__get();

	datum = (c = ((tag & 8) ? p_bcastGet8u() : 0));
	if (tag & 4) c = p_bcastGet8u();
	datum = ((datum << 8) | c);
	if (tag & 2) c = p_bcastGet8u();
	datum = ((datum << 8) | c);
	if (tag & 1) c = p_bcastGet8u();
	return((datum << 8) | c);
}


p_64u
p_bcastPut64u(register p_64u datum)
{
	p_bcastPut32u((p_32u) (datum >> 32));
	p_bcastPut32u((p_32u) datum);
	return(datum);
}


p_64u
p_bcastGet64u(void)
{
	register p_64u result = p_bcastGet32u();
	return((result << 32) | p_bcastGet32u());
}


#define	P_DECLARE(suffix) \
static p_##suffix p_bcastCache##suffix = 0; \
 \
p_##suffix \
p_bcastPut##suffix(register p_##suffix datum) \
{ \
	register p_##suffix xdatum = (datum ^ p_bcastCache##suffix); \
	p_bcastPut##suffix##u((p_##suffix##u) xdatum); \
	return(p_bcastCache##suffix = datum); \
} \
 \
p_##suffix \
p_bcastGet##suffix() \
{ \
	register p_##suffix xdatum = ((p_##suffix) p_bcastGet##suffix##u()); \
	return(p_bcastCache##suffix = (xdatum ^ p_bcastCache##suffix)); \
}

P_DECLARE(16)
P_DECLARE(32)
P_DECLARE(64)


#undef	P_DECLARE

static p_32u p_bcastCacheptr = 0;

p_ptr
p_bcastPutptr(register p_ptr datum)
{
	register p_32u xdatum = (p_bcastCacheptr ^ ((p_32u) datum));
	p_bcastPut32u(xdatum);
	p_bcastCacheptr = ((p_32u) datum);
	return(datum);
}


p_ptr
p_bcastGetptr(void)
{
	register p_32u datum = (p_bcastCacheptr ^ p_bcastGet32u());
	return((p_ptr) (p_bcastCacheptr = datum));
}

/*	p_bcastPutz() and p_bcastGetz()...

	We have already developed a very fancy hash-indexed, cached,
	run-length encoded, compression scheme...  but this does not
	use it.  We will distribute that version when we have more
	information about the type of data these routines are used to
	transmit, so that we can better tune the hashing, etc.

	For now, these routines simply use a run-length encoding
	similar to the compression used to send p_32u objects.
*/

inline void
p_bcastPutz(register void *source,
register int bytes)
{
	register p_8u prev = 0;

	while (bytes >= 4) {
		/* Use simple compression algorithm...
		   4-bit tag says which bytes differ
		*/
		register p_8u c0;
		register p_8u c1;
		register p_8u c2;
		register p_8u c3;
		register int tag;

		c0 = *((p_8u *) source);
		++source;
		c1 = *((p_8u *) source);
		++source;
		c2 = *((p_8u *) source);
		++source;
		c3 = *((p_8u *) source);
		++source;
		tag = (((c0 == prev) ? 0 : 1) |
		       ((c1 == c0) ? 0 : 2) |
		       ((c2 == c1) ? 0 : 4) |
		       ((c3 == c2) ? 0 : 8));
		p__put(tag);
		if (tag & 1) p_bcastPut8u(c0);
		if (tag & 2) p_bcastPut8u(c1);
		if (tag & 4) p_bcastPut8u(c2);
		if (tag & 8) p_bcastPut8u(c3);
		prev = c3;
		bytes -= 4;
	}

	while (bytes-- > 0) {
		p_bcastPut8u(*((p_8u *) source));
		++source;
	}
}


inline void
p_bcastGetz(register void *dest,
register int bytes)
{
	register p_8u prev = 0;

	while (bytes >= 4) {
		/* Use simple compression algorithm...
		   4-bit tag says which bytes differ
		*/
		register p_8u c0;
		register p_8u c1;
		register p_8u c2;
		register int tag = p__get();

		if (tag & 1) prev = p_bcastGet8u();
		c0 = prev;
		if (tag & 2) prev = p_bcastGet8u();
		c1 = prev;
		if (tag & 4) prev = p_bcastGet8u();
		c2 = prev;
		if (tag & 8) prev = p_bcastGet8u();
		if (dest) {
			*((p_8u *) dest) = c0;
			++dest;
			*((p_8u *) dest) = c1;
			++dest;
			*((p_8u *) dest) = c2;
			++dest;
			*((p_8u *) dest) = prev;
			++dest;
		}
		bytes -= 4;
	}

	if (dest) {
		while (bytes-- > 0) {
			*((p_8u *) dest) = p_bcastGet8u();
			++dest;
		}
	} else while (bytes-- > 0) p_bcastGet8u();
}

/*	p_bcastPut/Get() for 32f and 64f...

	Really can't easily compress floats; if you think about it, a
	floating point number has already undergone lossy compression.
	We'll use the compressed block transmission code.  Notice that
	this is not usable with a heterogeneous mix of floating point
	formats.
*/

static p_32f p_temp32f;
static p_64f p_temp64f;

p_32f
p_bcastPut32f(register p_32f datum)
{
	p_temp32f = datum;
	p_bcastPutz(((void *) &p_temp32f), sizeof(p_32f));
	return(datum);
}


p_32f
p_bcastGet32f(void)
{
	p_bcastGetz(((void *) &p_temp32f), sizeof(p_32f));
	return(p_temp32f);
}


p_64f
p_bcastPut64f(register p_64f datum)
{
	p_temp64f = datum;
	p_bcastPutz(((void *) &p_temp64f), sizeof(p_64f));
	return(datum);
}


p_64f
p_bcastGet64f(void)
{
	p_bcastGetz(((void *) &p_temp64f), sizeof(p_64f));
	return(p_temp64f);
}

/*	p_putGet() functions...

	The basic algorithm is to perform a sequence of broadcasts,
	however, this sequence skips broadcasts from PEs whose values
	will not be used by any processor.
*/

p_1u
p_putGet1u(register p_1u datum,
register int source)
{
	return((p__waitbar(datum) & (1 << source)) ? 1 : 0);
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_putGet##suffix(register p_##suffix datum, \
register int source) \
{ \
	register p_##suffix d = datum; \
	register int who = 0; \
	register p_bar mask = \
		p_reduceOrbar((source == IPROC) ? 0 : (1 << source)); \
 \
	do { \
		if (mask & 1) { \
			if (IPROC == who) { \
				/* Our turn to send data */ \
				p_bcastPut##suffix(datum); \
			} else if (source == who) { \
				/* We just get and store data... */ \
				d = p_bcastGet##suffix(); \
			} else { \
				/* Just be there for whoever... */ \
				p_bcastGet##suffix(); \
			} \
		} \
		++who; \
		mask >>= 1; \
	} while (who < NPROC); \
 \
	return(d); \
}

P_DECLARE(bar)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)
P_DECLARE(ptr)

#undef	P_DECLARE

void
p_putGetz(register void *dest,
register void *src,
register int n,
register int source)
{
	register int who = 0;
	register p_bar mask =
		p_reduceOrbar(((source == IPROC) ||
			       (dest == ((void *) 0))) ?
			       0 :
			       (1 << source));

	do {
		if (mask & 1) {
			if (IPROC == who) {
				/* Our turn to send data */
				p_bcastPutz(src, n);
			} else if (source == who) {
				/* We just get and store data... */
				p_bcastGetz(dest, n);
			} else {
				/* Just be there for whoever... */
				p_bcastGetz(((void *) 0), n);
			}
		}
		++who;
		mask >>= 1;
	} while (who < NPROC);

	if ((source == IPROC) && dest) {
		memcpy(dest, src, n);
	}
}

/*	p_gather() functions...
*/
#define	P_DECLARE(suffix) \
inline void \
p_gather##suffix(register p_##suffix *dest, \
register p_##suffix datum) \
{ \
	register p_##suffix d; \
	register int who = 0; \
 \
	do { \
		if (IPROC == who) { \
			/* Our turn to send and store data */ \
			p_bcastPut##suffix(d = datum); \
		} else { \
			/* We just get and store data... */ \
			d = p_bcastGet##suffix(); \
		} \
 \
		if (dest) dest[who] = d; \
	} while (++who < NPROC); \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_gatherz(register void *dest,
register void *src,
register int n)
{
	register int who = 0;

	do {
		if (IPROC == who) {
			/* Our turn to send and store data */
			p_bcastPutz(src, n);
			if (dest) {
				memcpy(dest, src, n);
			}
		} else {
			/* We just get and store data... */
			p_bcastGetz(dest, n);
		}

		if (dest) dest += n;
	} while (++who < NPROC);
}


/*	p_scatPut() and p_scatGet() functions...

	This is really a very simple operation to construct using the
	other routines, but it has been added to the library by
	popular demand....
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scatPut##suffix(register p_##suffix *src) \
{ \
	register int who = 0; \
 \
	do { \
		p_bcastPut##suffix(src[who]); \
	} while (++who < NPROC); \
 \
	return(src[IPROC]); \
} \
 \
p_##suffix \
p_scatGet##suffix(void) \
{ \
	register p_##suffix t; \
	register p_##suffix result = 0; \
	register int who = 0; \
 \
	do { \
		t = p_bcastGet##suffix(); \
		if (who == IPROC) result = t; \
	} while (++who < NPROC); \
 \
	return(result); \
}

P_ALPHA

#undef	P_DECLARE

void
p_scatPutz(register void *dest,
register void *src,
register int n)
{
	register int who = 0;

	do {
		p_bcastPutz(src, n);
		if (dest && (who == IPROC)) {
			memcpy(dest, src, n);
		}
		src += n;
	} while (++who < NPROC);
}

void
p_scatGetz(register void *dest,
register int n)
{
	register int who = 0;

	do {
		p_bcastGetz(((who == IPROC) ? dest : ((void *) 0)), n);
	} while (++who < NPROC);
}

/*	p_all2all() functions...

	The meaning of all2all here is essentially the same as that
	used in MPI, namely, every PE has different data for each PE.
	This is more properly called "personalized all-to-all," but so
	much for terminology....  If our hardware supports putGet,
	then NPROC-1 putGet should be used; TTL_PAPERS does not
	support putGet directly, but it is close enough....
*/
#define	P_DECLARE(suffix) \
inline void \
p_all2all##suffix(register p_##suffix *dest, \
register p_##suffix *src) \
{ \
	register int dist = 1; \
 \
	if (dest) { \
		dest[IPROC] = src[IPROC]; \
		do { \
			register int to = (IPROC + dist); \
			register int from = (IPROC - dist); \
 \
			if (to >= NPROC) to -= NPROC; \
			if (from < 0) from += NPROC; \
			dest[from] = p_putGet##suffix(src[to], from); \
		} while (++dist < NPROC); \
	} else { \
		do { \
			register int to = (IPROC + dist); \
 \
			if (to >= NPROC) to -= NPROC; \
			p_putGet##suffix(src[to], NPROC); \
		} while (++dist < NPROC); \
	} \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_all2allz(register void *dest,
register void *src,
register int n)
{
	register int dist = 1;

	if (dest) {
		register int off = (n * IPROC);

		memcpy(dest + off, src + off, n);
		do {
			register int to = (IPROC + dist);
			register int from = (IPROC - dist);

			if (to >= NPROC) to -= NPROC;
			if (from < 0) from += NPROC;
			p_putGetz(dest + (from * n),
				  src + (to * n),
				  n,
				  from);
		} while (++dist < NPROC);
	} else {
		do {
			register int to = (IPROC + dist);

			if (to >= NPROC) to -= NPROC;
			p_putGetz(((void *) 0),
				  src + (to * n),
				  n,
				  IPROC);
		} while (++dist < NPROC);
	}
}
#endif /* if NPROC > 1 */
