/*	scan.c

	AFAPI scan (parallel prefix) function definitions for
	TTL_PAPERS.

	July 1996 by H. Dietz
*/
#if	NPROC > 1

#include "intern.h"

/*	p_scanAnd() functions...

	Sneaky implementation does at most NPROC-1 broadcasts and
	detects when result is 0 to truncate the broadcast sequence.
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanAnd##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix product = p_##suffix##_ALL1; \
	register p_##suffix result = 0; \
 \
	do { \
		if (who == IPROC) { \
			result = (product &= datum); \
			p_bcastPut##suffix(product); \
		} else { \
			product = p_bcastGet##suffix(); \
		} \
		if (product == 0) return(result); \
	} while (++who < (NPROC - 1)); \
 \
	return((IPROC == (NPROC - 1)) ? (product & datum) : result); \
}

P_DECLARE(1u)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE


p_ptr
p_scanAndptr(register p_ptr datum)
{
	return((p_ptr) p_scanAnd32u((p_32u) datum));
}

/*	p_scanOr() functions...

	Sneaky implementation does at most NPROC-1 broadcasts and
	detects when result is ~0 to truncate the broadcast sequence.
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanOr##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix sum = 0; \
	register p_##suffix result = p_##suffix##_ALL1; \
 \
	do { \
		if (who == IPROC) { \
			result = (sum |= datum); \
			p_bcastPut##suffix(sum); \
		} else { \
			sum = p_bcastGet##suffix(); \
		} \
		if (sum == p_##suffix##_ALL1) return(result); \
	} while (++who < (NPROC - 1)); \
 \
	return((IPROC == (NPROC - 1)) ? (sum | datum) : result); \
}

P_DECLARE(1u)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE


p_ptr
p_scanOrptr(register p_ptr datum)
{
	return((p_ptr) p_scanOr32u((p_32u) datum));
}

/*	p_scanXor() functions...

	Implementation does NPROC-1 broadcasts.
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanXor##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix sum = 0; \
	register p_##suffix result = 0; \
 \
	do { \
		if (who == IPROC) { \
			result = (sum ^= datum); \
			p_bcastPut##suffix(sum); \
		} else { \
			sum = p_bcastGet##suffix(); \
		} \
	} while (++who < (NPROC - 1)); \
 \
	return((IPROC == (NPROC-1)) ? (sum ^ datum) : result); \
}

P_DECLARE(1u)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE


p_ptr
p_scanXorptr(register p_ptr datum)
{
	return((p_ptr) p_scanXor32u((p_32u) datum));
}

/*	p_scanMin() functions...

	Very clever implementation outputs the PE0 value and then
	votes to see if any of the next four PEs would have a new
	minimum.  If one of the next four PEs does, then that PE will
	output its value. Otherwise, we continue the voting with the
	next four PEs.
*/

p_1u
p_scanMin1u(register p_1u datum)
{
	return(p_scanAnd1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMin##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register int pos; \
	register int mask; \
	register p_##suffix min = datum; \
	register p_##suffix result = 0; \
 \
newmin: \
	if (who == (NPROC-1)) { \
		if (who == IPROC) { \
			result = ((datum < min) ? datum : min); \
		} \
		return(result); \
	} \
 \
	if (who == IPROC) { \
		result = (min = datum); \
		p_bcastPut##suffix(datum); \
	} else { \
		min = p_bcastGet##suffix(); \
	} \
newgroup: \
	pos = (IPROC - ++who); \
	mask = 0xf; \
	if ((pos >= 0) && (pos < 4)) { \
		if (datum < min) { \
			mask = (0xf ^ (1 << pos)); \
		} else { \
			result = min; \
		} \
	} \
	mask = p__nand(mask); \
	if (mask & 1) goto newmin; \
	++who; \
	if (mask & 2) goto newmin; \
	++who; \
	if (mask & 4) goto newmin; \
	++who; \
	if (mask & 8) goto newmin; \
	if (who < (NPROC-1)) goto newgroup; \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanMax() functions...

	Done like p_scanMin().
*/

p_1u
p_scanMax1u(register p_1u datum)
{
	return(p_scanOr1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMax##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register int pos; \
	register int mask; \
	register p_##suffix max = datum; \
	register p_##suffix result = 0; \
 \
newmax: \
	if (who == (NPROC-1)) { \
		if (who == IPROC) { \
			result = ((datum > max) ? datum : max); \
		} \
		return(result); \
	} \
 \
	if (who == IPROC) { \
		result = (max = datum); \
		p_bcastPut##suffix(datum); \
	} else { \
		max = p_bcastGet##suffix(); \
	} \
newgroup: \
	pos = (IPROC - ++who); \
	mask = 0xf; \
	if ((pos >= 0) && (pos < 4)) { \
		if (datum > max) { \
			mask = (0xf ^ (1 << pos)); \
		} else { \
			result = max; \
		} \
	} \
	mask = p__nand(mask); \
	if (mask & 1) goto newmax; \
	++who; \
	if (mask & 2) goto newmax; \
	++who; \
	if (mask & 4) goto newmax; \
	++who; \
	if (mask & 8) goto newmax; \
	if (who < (NPROC-1)) goto newgroup; \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanAdd() functions...

	Do the obvious, N-1 transmission, thing.
*/

p_1u
p_scanAdd1u(register p_1u datum)
{
	return(p_scanOr1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanAdd##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix sum = 0; \
	register p_##suffix result = 0; \
 \
	do { \
		if (who == IPROC) { \
			result = (sum += datum); \
			p_bcastPut##suffix(sum); \
		} else { \
			sum = p_bcastGet##suffix(); \
		} \
	} while (++who < (NPROC - 1)); \
 \
	return(P_CAST_##suffix((IPROC==(NPROC-1)) ? (sum + datum) : result)); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)


#undef P_DECLARE

/*	p_scanMul() functions...

	Like p_scanAdd(), do the obvious, N-1 transmission, thing.
	However, stop everything if product ever becomes 0, because
	future multiplies can't change that value.
*/

p_1u
p_scanMul1u(register p_1u datum)
{
	return(p_scanAnd1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMul##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix product = 1; \
	register p_##suffix result = 0; \
 \
	do { \
		if (who == IPROC) { \
			result = (product *= datum); \
			p_bcastPut##suffix(product); \
		} else { \
			product = p_bcastGet##suffix(); \
		} \
		if (product == 0) return(result); \
	} while (++who < (NPROC - 1)); \
 \
	return(P_CAST_##suffix((IPROC == (NPROC - 1)) ? \
		 (product * datum) : result)); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef P_DECLARE
#endif /* if NPROC > 1 */
