/*	signal.c

	AFAPI signal definitions for TTL_PAPERS.

	July 1996 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"


#define	S_SIGMAX	3


static p_8u s_outlast;

/*	Pre-defined signal handlers...
*/

void
s_ignore(register p_bar m)
{
	/* Nothing to do...  that's why it's called s_ignore() */
}

/*	s_vector[] array...

	The following is the magic internal structure used to vector
	signals to the correct handlers.  It is specific to TTL_PAPERS.
*/

static void (*s_vector[S_SIGMAX])(register p_bar mask) = {
	/* Normal signals... */
	s_ignore,
	s_ignore,
	s_ignore
};

/*	s_install()

	Install a signal handler...
*/

int
s_install(register int s,
register void (*f)(p_bar m))
{
	/* Install the signal handler f()... */
	register int which = s;

	/* If s is S_SIGANY, pick one that is s_ignore()... */
	if (which == S_SIGANY) {
		for (which = 0; which<S_SIGMAX; ++which) {
			if (s_vector[which] == s_ignore) break;
		}
	}

	/* Got a valid signal number? */
	if ((which < 0) || (which >= S_SIGMAX)) {
		/* Nope.  Return with error. */
		return(-1);
	}

	/* Ok, install it... */
	s_vector[which] = f;

	/* Return signal number */
	return(which);
}

/*	s_init() and s_exit()...
*/

int
s_init(void)
{
	register unsigned char t;
	register unsigned char match;

	/* Flip the strobe bits and signal idle... */
	t = ((p_outlast ^ (0x80 | 0x40)) & 0xf0);
	t |= (s_outlast = 0x07);
	p__outb(t, p_outport);
	p_outlast = t;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t, p_outport);
		}
	}

	/* Ok, it worked...? */
	return(0);
}


void
s_exit(void)
{
	/* Tell everybody we are idle and want to exit... */
	register p_8u t = ((p_outlast & 0xf0) | (s_outlast = 0x0f));
	register p_8u pin;
	p__outb(t, p_outport);
	p_outlast = t;

	/* Wait for everybody to agree that the signal system
	   should be exited, polling as we wait
	*/
	do {
		pin = p__inb(p_inport);

		if (pin & 0x38) {
			s_signal(S_SIGANY);
		}
	} while (pin & 0x40);

	/* To be sure everyone knows, do a real barrier... */
	p__wait();
}

/*	s_signal()...

	This code must recognize who has sent signals, create masks grouping
	similar signals together, and invoke the appropriate handler
	functions in a way that gives s_exit the lowest priority....

	There are only 4 bits in the TTL_PAPERS data path, and we use
	each bit to define a signal, thus, we have only four signals.
	In fact, one of those signals is reserved for disabling the
	signal system, so only three are available for users.  Signal
	number s is bit (1 << s).

	Later, these 3 signals will probably be expanded by using a
	two-level decoding process for the third signal to yield a
	total of 6 signals, the first 2 of which are "fast."  The
	other choice would be to make 9 user signals with only the
	first being "fast."  Anyway, for now there are 3
	"fast" user signals....
*/

void
s_signal(register int s)
{
	register unsigned char t, t2;
	register unsigned char match;
	register int result;

	/* A valid signal...? If so, update s_oulast... */
	s_outlast = (((s >= 0) && (s <= S_SIGMAX)) ?
		     (s_outlast & ~(1 << s)) :
		     (s_outlast | 0x07));

	/* Paste-in the signal, flip strobe bits, and output it */
	t2 = ((p_outlast & 0xf0) | s_outlast);
	t = (t2 ^ (0x80 | 0x40));
	p__outb(t, p_outport);
	p_outlast = t2;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			/* The LED will be turned green on the next p__outb */
		}
	}

	/* Read the data value again (time for data to settle) */
	result = p__inb(p_inport);

	/* Now do another barrier to say everyone has seen the data
	   value.  Notice that we save a little overhead by reusing
	   the match value from above....
	   Do not mess with the red LED since it is highly unlikely
	   we would get stuck right after the above barrier.
	*/
	p__outb(t2, p_outport);
	while ((p__inb(p_inport) & 0x80) != match) ;

	/* Result is now a mask saying which signals...  process them
	   in priority order, shared memory first.  Notice that higher
	   bit positions are lower priority, with disabling the signal
	   system lowest.
	*/
	if (result & 0x08) {
		register p_bar mask = p_waitbar(s == 0);
		s_vector[0](mask);
	}
	if (result & 0x10) {
		register p_bar mask = p_waitbar(s == 1);
		s_vector[1](mask);
	}
	if (result & 0x20) {
		register p_bar mask = p_waitbar(s == 2);
		s_vector[2](mask);
	}


	/* Done...  re-enable signals... */
	t2 = ((p_outlast & 0xf0) | (s_outlast |= 0x07));
	t = (t2 ^ (0x80 | 0x40));
	p__outb(t, p_outport);
	p_outlast = t2;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			/* The LED will be turned green on the next p__outb */
		}
	}

	/* Don't bother reading...  but do another barrier to be
	   certain that this value has had lots of settle time....
	   Do not mess with the red LED since it is highly unlikely
	   we would get stuck right after the above barrier.
	*/
	p__outb(t2, p_outport);
	while ((p__inb(p_inport) & 0x80) != match) ;
}

/*	Inlined signal stuff...

	These are just callable versions of the simple interface
	routines....
*/

p_1u
s_pending(void)
{
	return(s__pending());
}

void
s_poll(void)
{
	s__poll();
}

void
s_wait(void)
{
	s__wait();
}

#endif /* if NPROC > 1 */
