/*	vwfb.c

	Video Wall Frame Buffer support.

	March 1999, Hank Dietz
*/


#include <AFAPI/vwsup.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <malloc.h>

#include <linux/fb.h>


static unsigned short *frame_buf;
static void *map;
static int fd;

#ifdef	VWTURN_OFF_CURSOR
#ifdef	TURN_OFF_FB_CURSOR
static struct fb_cursorstate cursor_state;
#endif
#endif


void
vwbegin(void)
{
	struct fb_var_screeninfo si;

	/* First, open the file... */
	if ((fd = open("/dev/fb0", O_RDWR)) < 0) {
		fprintf(stderr, "Cannot open framebuffer /dev/fb0\n");
		exit(1);
	}

	if (ioctl(fd, FBIOGET_VSCREENINFO, &si)) {
		fprintf(stderr, "Cannot get framebuffer screen info.\n");
		exit(1);
	}
	vwpanel.dim.x = si.xres;
	vwpanel.dim.y = si.yres;
	vwpanel.depth = si.bits_per_pixel;
	vwpanel.bytes = si.xres * si.yres * (si.bits_per_pixel / 8);
	if (vwpanel.depth & 0x7) { /* pixels must be an integral number of bytes in size */
		fprintf(stderr,
			"Framebuffer bitdepth (%d) not supported\n",
			vwpanel.depth);
		exit(1);
	}

	if ((vwpanel.dim.x > VWWIDE) ||
	    (vwpanel.dim.y > VWTALL)) {
		fprintf(stderr,
"Recompile Video Wall support with VWWIDE>=%d and VWTALL>=%d\n",
			vwpanel.dim.x,
			vwpanel.dim.y);
		exit(1);
	}

	/* Map the active region of the frame buffer. */
	map = ((short *) mmap(0,		/* Put it anywhere */
			     vwpanel.bytes,	/* Map the whole file */
			     (PROT_READ | PROT_WRITE),
			     MAP_SHARED,	/* Not just for me */
			     fd,		/* The file */
			     0));		/* Right from the start */
	if (((int) map) == -1) {
		fprintf(stderr, "Cannot map framebuffer /dev/fb0\n");
		exit(1);
	}

	frame_buf = map;

	vwpanel.dirty.min.x = -1;
	vwpanel.dirty.min.y = -1;
	vwpanel.dirty.max.x = -1;
	vwpanel.dirty.max.y = -1;
	vwpanel.data.data = malloc(vwpanel.bytes);

#ifdef	VWTURN_OFF_CURSOR
#ifdef	TURN_OFF_FB_CURSOR
	{
		struct fb_cursorstate cs;

		/* Get & save current cursor state */
		if (ioctl(fd, FBIOGET_CURSORSTATE, &cursor_state)) {
			fprintf(stderr, "Cannot get cursor state info.\n");
			exit(1);
		}
		/* Turn off cursor */
		cs.xoffset = cursor_state.xoffset;
		cs.yoffset = cursor_state.yoffset;
		cs.mode = FB_CURSOR_OFF;
		if (ioctl(fd, FBIOPUT_CURSORSTATE, &cs)) {
			fprintf(stderr, "Cannot set cursor state.\n");
			exit(1);
		}
	}
#else
	system("/usr/bin/setterm -cursor off");
#endif
#endif

	vpclear(&vwpanel);
}

inline void
vppaint(register vwbuf *panel)
{
	/* Write vwbuf to display in one blit */
	memcpy(frame_buf,
	       panel->data.data,
	       panel->bytes);
}

void
vwpaint(register vwwall *wall)
{
	/* vwpaint(0) paints my panel no matter what;
	   vwpaint(wall) paints my panel only if it's in wall
	*/
	if (((!wall) && vwspace[IPROC]) ||
	    (wall == vwspace[IPROC])) {
		vppaint(&vwpanel);
	}
}

void
vwend(void)
{
	if (vwspace[IPROC] && vwpanel.data.data) {
		free(vwpanel.data.data);
	}
#ifdef	VWTURN_OFF_CURSOR
	/* Restore cursor state */
#ifdef	TURN_OFF_FB_CURSOR
	if (ioctl(fd, FBIOPUT_CURSORSTATE, &cursor_state)) {
		fprintf(stderr, "Cannot restore FB cursor state.\n");
	}
#else
	system("/usr/bin/setterm -cursor on");
#endif
#endif

	close(fd);
	munmap(map, vwpanel.bytes);
}

