/*	vwkeybrd.c

	Video Wall KEYBoaRD handlers.

	For now, this code uses standard UNIX facilities for
	raw-mode input from /dev/tty.  This makes for a crude,
	but highly portable, interface.

	March 1999, H. Dietz
*/


#include <AFAPI/vwsup.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <termios.h>


static int keybrdfd = -1;
static struct termios keybrdsaved;

void
vwbeginKeybrd(void)
{
	if (IPROC == CPROC) {
		keybrdfd = open("/dev/console", (O_RDONLY | O_NDELAY));
		if (keybrdfd >= 0) {
			struct termios t;
			tcgetattr(keybrdfd, &t);
			keybrdsaved = t;
			cfmakeraw(&t);
			tcsetattr(keybrdfd, 0, &t);
		}
	}
}

extern int
vwreadKeybrd(register vwtime until)
{
	if (IPROC == CPROC) {
        	if (keybrdfd < 0) return(p_bcastPut32(VWNOKEYBRD));

		do {
			/* use select() to check for input */
			fd_set fds, nul;
			struct timeval t;
			t.tv_sec = 0;
			t.tv_usec = 0;
			FD_ZERO(&nul);
			FD_ZERO(&fds);
			FD_SET(keybrdfd, &fds);
			select(keybrdfd+1, &fds, &nul, &nul, &t);
			if (FD_ISSET(keybrdfd, &fds)) {
				/* something is there */
				unsigned char c;

				if (read(keybrdfd, &c, 1) != 1) {
					return(p_bcastPut32(VWEOF));
				} else {
					/* note: chars are in 0..255 */
					return(p_bcastPut32(c));
				}
			}
		} while ((until != 0.0) && (vwnow() < until));
		return(p_bcastPut32(VWNOKEY));
	} else {
	        return(p_bcastGet32());
	}
}

void
vwendKeybrd(void)
{
	if (IPROC == CPROC) {
		if (keybrdfd >= 0) {
			tcsetattr(keybrdfd, 0, &keybrdsaved);
			close(keybrdfd);
			keybrdfd = -1;
		}
	}
}
