/*	vwmag.c

	Video Wall MAGnify routines.

	February 1999 by H. Dietz
*/


#include <AFAPI/vwsup.h>
#include <math.h>


void
vwscale(register vwwall *destwall,
register vwfbox destarea,
register vwfxy destctr,
register vwbuf *srcimage,
register vwfxy srcctr,
register vwfxy srcmag,
register int frac)
{
	register vwfbox srcarea, refarea;

	/* If destwall is 0 or doesn't contain my panel,
	   it isn't my problem
	*/
	if (!destwall) destwall = vwspace[IPROC];
	if ((!destwall) ||
	    (destwall != vwspace[IPROC])) {
		return;
	}

	/* Compute the virtual image position */
	srcctr.x = 1.0 - srcctr.x;
	srcctr.y = 1.0 - srcctr.y;
	srcarea.min.x = destctr.x - (srcmag.x * (1.0 - srcctr.x));
	srcarea.min.y = destctr.y - (srcmag.y * (1.0 - srcctr.y));
	srcarea.max.x = destctr.x + (srcmag.x * srcctr.x);
	srcarea.max.y = destctr.y + (srcmag.y * srcctr.y);
	refarea = srcarea;

	/* Now clip it */
	if (srcarea.min.x < destarea.min.x) {
		srcarea.min.x = destarea.min.x;
	} else if (srcarea.min.x > destarea.min.x) {
		destarea.min.x = srcarea.min.x;
	}
	if (srcarea.min.y < destarea.min.y) {
		srcarea.min.y = destarea.min.y;
	} else if (srcarea.min.y > destarea.min.y) {
		destarea.min.y = srcarea.min.y;
	}
	if (srcarea.max.x > destarea.max.x) {
		srcarea.max.x = destarea.max.x;
	} else if (srcarea.max.x < destarea.max.x) {
		destarea.max.x = srcarea.max.x;
	}
	if (srcarea.max.y > destarea.max.y) {
		srcarea.max.y = destarea.max.y;
	} else if (srcarea.max.y < destarea.max.y) {
		destarea.max.y = srcarea.max.y;
	}

	/* Now rescale srcarea by refarea */
	srcarea.min.x = ((srcarea.min.x - refarea.min.x) / srcmag.x);
	srcarea.min.y = ((srcarea.min.y - refarea.min.y) / srcmag.y);
	srcarea.max.x = ((srcarea.max.x - refarea.min.x) / srcmag.x);
	srcarea.max.y = ((srcarea.max.y - refarea.min.y) / srcmag.y);

	vwres(destwall,
	      destarea,
	      srcimage,
	      srcarea,
	      frac);
}

float
vpaspect(register vwbuf *panel)
{
	if (panel->dim.y == 0) return(FLT_MAX);
	return(((float) panel->dim.x) / panel->dim.y);
}

float
vwaspect(register vwwall *destwall,
register vwfbox destarea)
{
	register float x;
	register float y;

	if (destwall == 0) {
		if ((destwall = vwspace[IPROC]) == 0) {
			return(1.0);
		}
	}
	x = ((destwall->all.max.x - destwall->all.min.x) *
	     (destarea.max.x - destarea.min.x));
	y = ((destwall->all.max.y - destwall->all.min.y) *
	     (destarea.max.y - destarea.min.y));
	return((y == 0) ? FLT_MAX : (x / y));
}

void
vwmag(register vwwall *destwall,
register vwfbox destarea,
register vwfxy destctr,
register vwbuf *srcimage,
register vwfxy srcctr,
register float srcmag,
register int frac)
{
	register float srca = vpaspect(srcimage);
	register float desta = vwaspect(destwall, destarea);
	register vwfxy srcmag2;

	if (srca > desta) {
		srcmag2.x = srcmag;
		srcmag2.y = ((desta * srcmag) / srca);
	} else {
		srcmag2.x = ((srca * srcmag) / desta);
		srcmag2.y = srcmag;
	}

	vwscale(destwall,
		destarea,
		destctr,
		srcimage,
		srcctr,
		srcmag2,
		frac);
}
