/*	vwmouse.c

	Video Wall MOUSE handlers.

	This code uses the vgamouse routines, which are quite
	generic.  However, it is NOT used with the X versions of
	the video wall, since X handles the mouse itself.

	February 1999, H. Dietz
*/


#include <AFAPI/vwsup.h>
#include <vgamouse.h>


void
vpbeginMouse(void)
{
	/* May need to set other VGA mouse parameters */
	mouse_init(VW_MOUSE_DEV,
		   VW_MOUSE_TYPE,
		   MOUSE_DEFAULTSAMPLERATE);
	mouse_setxrange(0, VW_MOUSE_XRES);
	mouse_setyrange(0, VW_MOUSE_YRES);
	mouse_setwrap(VW_MOUSE_WRAP);
}

static	vwinfoMouse prevMouse = { { 0.0, 0.0 }, 0, 0, 0 };

vwinfoMouse
vpreadMouse(register vwtime until)
{
	vwinfoMouse retval;

	do {
		mouse_update();
		retval.pos.x = (mouse_getx() *
				(1.0 / VW_MOUSE_XRES));
		retval.pos.y = (mouse_gety() *
				(1.0 / VW_MOUSE_YRES));
		retval.left = 0;
		retval.center = 0;
		retval.right = 0;
		switch (mouse_getbutton() &
			(MOUSE_LEFTBUTTON |
			 MOUSE_RIGHTBUTTON)) {
		case MOUSE_LEFTBUTTON:
			retval.left = 1;
			break;
		case MOUSE_RIGHTBUTTON:
			retval.right = 1;
			break;
		case (MOUSE_LEFTBUTTON |
		      MOUSE_RIGHTBUTTON):
			retval.center = 1;
		}
	} while ((until != 0.0) &&
		 (vwnow() < until) &&
		 (prevMouse.pos.x == retval.pos.x) &&
		 (prevMouse.pos.y == retval.pos.y) &&
		 (prevMouse.left == retval.left) &&
		 (prevMouse.center == retval.center) &&
		 (prevMouse.right == retval.right));

	prevMouse = retval;
	return(retval);
}

void
vpendMouse(void)
{
	mouse_close();
}

void
vwbeginMouse(void)
{
	if (IPROC == CPROC)
		vpbeginMouse();
}

vwinfoMouse
vwreadMouse(register vwtime until)
{
	vwinfoMouse t;
	if (IPROC == CPROC) {
		t = vpreadMouse(until);
		p_bcastPutz(&t, sizeof(t));
	} else {
		p_bcastGetz(&t, sizeof(t));
	}
	return(t);
}

void
vwendMouse(void)
{
	if (IPROC == CPROC)
		vpendMouse();
}
