/*	vwres.c

	Video Wall RESample routines.

	February 1999 by H. Dietz
*/


#include <stdio.h>
#include <stdlib.h>
#include <AFAPI/vwsup.h>


static unsigned char vwbias[VWWIDE * VWTALL];
static unsigned char vwdither[3 * VWWIDE * VWTALL];
static int initdone = 0;


static void
resinit(void)
{
	register unsigned char *p;

	if (initdone) return;

	p = &(vwbias[0]);
	do {
		*p = ((rand() >> 3) & 0xff);
	} while (++p < &(vwbias[VWWIDE * VWTALL]));

	p = &(vwdither[0]);
	do {
		*p = ((rand() >> 3) & 0037);
		*++p = ((rand() >> 3) & 0037);
		*++p = ((rand() >> 3) & 0077);
	} while (++p < &(vwdither[3 * VWWIDE * VWTALL]));

	initdone = 1;
}


extern inline void
recolor8(register unsigned char *p,
register unsigned char *base,
register int r,
register int g,
register int b)
{
	register unsigned char *q = &(vwdither[3 * (p - base)]);

	r &= 0xff;
	g &= 0xff;
	b &= 0xff;
	if ((r += *q) > 255) r = 0007; else r = (((r) >> 5) & 0007);
	if ((g += *(++q)) > 255) g = 0070; else g = (((g) >> 2) & 0070);
	if ((b += *(++q)) > 255) b = 0300; else b = ((b) & 0300);
	*p = r + g + b;
}

extern inline void
recolor16(register unsigned short *p,
register unsigned short *base,
register int r,
register int g,
register int b)
{
	*p = ((((r) << 8) & 0xf800) |
	      (((g) << 3) & 0x07e0) |
	      (((b) >> 3) & 0x001f));
}

extern inline void
recolor24(register unsigned char *p,
register unsigned char *base,
register int r,
register int g,
register int b)
{
	*p = r;
	*(++p) = g;
	*(++p) = b;
}

extern inline void
recolor32(register unsigned int *p,
register unsigned int *base,
register int r,
register int g,
register int b)
{
	r &= 0xff;
	g &= 0xff;
	b &= 0xff;
	*p = ((r << 16) +
	      (g << 8) +
	      b);
}


static	int xoff[VWWIDE];
static	int yoff[VWTALL];
static	int xnxt[VWWIDE];
static	int ynxt[VWTALL];
static	unsigned short row0[VWWIDE * 3];
static	unsigned short row1[VWWIDE * 3];
static	unsigned char *rowdex[2];
static	int rownxt = 0;


static unsigned short *
resamp1d(register unsigned char *src,
register int width)
{
	register int i;
	register unsigned short *q;
	register unsigned short *dest;

	/* Is the src line we want already buffered? */
	if (rowdex[0] == src) { rownxt = 1; return(&(row0[0])); }
	if (rowdex[1] == src) { rownxt = 0; return(&(row1[0])); }

	/* No, select buffer for it */
	if (rownxt == 0) {
		rowdex[0] = src;
		dest = &(row0[0]);
		rownxt = 1;
	} else {
		rowdex[1] = src;
		dest = &(row1[0]);
		rownxt = 0;
	}
	q = dest;

	/* Compute and buffer this row */
	for (i=0; i<width; ++i) {
		register unsigned char *p = (src + xoff[i]);
		register int red = *p;
		register int green = *++p;
		register int blue = *++p;
		register int nfrac = xnxt[i];

		if (nfrac != 0) {
			/* Must merge with next x pixel */
			red = (((red << 16) +
				((*++p - red) * nfrac)) >> 8);
			green = (((green << 16) +
				  ((*++p - green) * nfrac)) >> 8);
			blue = (((blue << 16) +
				 ((*++p - blue) * nfrac)) >> 8);
		} else {
			red <<= 8;
			green <<= 8;
			blue <<= 8;
		}

		*q = red;
		*++q = green;
		*++q = blue;
		++q;
	}

	return(dest);
}


static int nxty;	/* bytes offset to next y */


/* Was "static unsigned char *" -- Tim 19991115 */
void
resamp(register int dxcrop,
register int dycrop,
register int sxdim,
register double sxoff,
register double sxcrop,
register int sydim,
register double syoff,
register double sycrop)
{
	/* Set-up for resampling image */
	register double off;
	register double scale;
	register int i;

	nxty = (3 * sxdim);	/* bytes offset to next y */

	/* Compute x offsets for resampling */
	off = (sxoff * (sxdim - 1));
	scale = ((sxcrop * (sxdim - 1)) / (dxcrop - 1));
	for (i=0; i<dxcrop; ++i) {
		xoff[i] = ((int) off);			/* rounds down */
		xnxt[i] = 0xffff * (off - xoff[i]);	/* 1/64k approx */
		xoff[i] *= 3;				/* pixel width */
		off += scale;
	}

	/* Compute y offsets for resampling */
	off = (syoff * (sydim - 1));
	scale = ((sycrop * (sydim - 1)) / (dycrop - 1));
	for (i=0; i<dycrop; ++i) {
		yoff[i] = ((int) off);			/* rounds down */
		ynxt[i] = 0xffff * (off - yoff[i]);	/* 1/64k approx */
		yoff[i] *= nxty;			/* row width */
		off += scale;
	}

	/* Clear row buffers */
	rownxt = 0;
	rowdex[1] = (rowdex[0] = ((unsigned char *) 0));
}


#define	RESAMP2D(bits, type) \
extern inline void \
vwres##bits(register unsigned type *dpbase, \
register int dxdim, \
register int dxoff, \
register int dxcrop, \
register int dydim, \
register int dyoff, \
register int dycrop, \
register unsigned char *spbase, \
register int sxdim, \
register double sxoff, \
register double sxcrop, \
register int sydim, \
register double syoff, \
register double sycrop, \
register int frac) \
{ \
	/* Resample and map to bits-deep color */ \
	register int i, j; \
	register unsigned type *r = \
		&(dpbase[(sizeof(type) / (bits / 8)) * \
			 (dxoff + (dyoff * dxdim))]); \
 \
	/* Take care of bias stuff */ \
	resinit(); \
 \
	/* Set-up resampling index and smoothing info */ \
	resamp(dxcrop, dycrop, sxdim, sxoff, sxcrop, sydim, syoff, sycrop); \
 \
	/* Now loop through y and x... */ \
	for (j=0; j<dycrop; ++j) { \
		register unsigned short *p = \
		  (resamp1d((spbase + yoff[j]), dxcrop) - 1); \
		register int nyfrac = ynxt[j]; \
		register unsigned char *biasp = \
		  &(vwbias[dxoff + ((dyoff + j) * dxdim)]); \
 \
		if (nyfrac) { \
			register unsigned short *q = \
			  (resamp1d((spbase + yoff[j] + nxty), dxcrop) - 1); \
 \
			for (i=0; i<dxcrop; ++i) { \
				if (frac >= biasp[i]) { \
					register int red = *++p; \
					register int green = *++p; \
					register int blue = *++p; \
 \
					red = (((red << 16) + \
						((*++q - red) * nyfrac)) >> 24); \
					green = (((green << 16) + \
						  ((*++q - green) * nyfrac)) >> 24); \
					blue = (((blue << 16) + \
						 ((*++q - blue) * nyfrac)) >> 24); \
 \
					recolor##bits(r, dpbase, red, green, blue); \
				} else { \
					p += 3; \
					q += 3; \
				} \
				r += (sizeof(type) / (bits / 8)); \
			} \
		} else { \
			for (i=0; i<dxcrop; ++i) { \
				if (frac >= biasp[i]) { \
					register int red = *++p >> 8; \
					register int green = *++p >> 8; \
					register int blue = *++p >> 8; \
 \
					recolor##bits(r, dpbase, red, green, blue); \
				} else { \
					p += 3; \
				} \
				r += (sizeof(type) / (bits / 8)); \
			} \
		} \
 \
		/* To the next row */ \
		r += ((sizeof(type) / (bits / 8)) * (dxdim - dxcrop)); \
	} \
}


RESAMP2D(8, char)
RESAMP2D(16, short)
RESAMP2D(24, char)
RESAMP2D(32, int)

#undef	RESAMP2D


inline void
vpres(register vwbuf *destpanel,
register vwbox destarea,
register vwbuf *srcimage,
register vwfbox srcarea,
register int frac)
{
	if (srcimage->depth != 24) {
		fprintf(stderr,
			"Only resampling from 24-bit images is supported\n");
		exit(1);
	}

	switch (destpanel->depth) {

#define	VPRES(bits) \
	case bits: \
		vwres##bits(destpanel->data.data##bits, \
			    destpanel->dim.x, \
			    destarea.min.x, \
			    (destarea.max.x - destarea.min.x), \
			    destpanel->dim.y, \
			    destarea.min.y, \
			    (destarea.max.y - destarea.min.y), \
			    srcimage->data.data24, \
			    srcimage->dim.x, \
			    srcarea.min.x, \
			    (srcarea.max.x - srcarea.min.x), \
			    srcimage->dim.y, \
			    srcarea.min.y, \
			    (srcarea.max.y - srcarea.min.y), \
			    frac); \
		break;

VPRES(8)
VPRES(16)
VPRES(24)
VPRES(32)

#undef	VPRES

	}

	/* Adjust the dirty region info... */
	if (destpanel->dirty.min.x < 0) {
		destpanel->dirty = destarea;
	} else {
		if (destarea.min.x < destpanel->dirty.min.x) {
			destpanel->dirty.min.x = destarea.min.x;
		}
		if (destarea.max.x > destpanel->dirty.max.x) {
			destpanel->dirty.max.x = destarea.max.x;
		}
		if (destarea.min.y < destpanel->dirty.min.y) {
			destpanel->dirty.min.y = destarea.min.y;
		}
		if (destarea.max.y > destpanel->dirty.max.y) {
			destpanel->dirty.max.y = destarea.max.y;
		}
	}
}


void
vwres(register vwwall *destwall,
register vwfbox destarea,
register vwbuf *srcimage,
register vwfbox srcarea,
register int frac)
{
	register float scale;
	vwfbox mypanel;
	vwbox d;

	/* If destwall is 0 or doesn't contain my panel,
	   it isn't my problem
	*/
	if (!destwall) destwall = vwspace[IPROC];
	if ((!destwall) ||
	    (destwall != vwspace[IPROC])) {
		return;
	}

	/* Scale mypanel coordinates to 0:1,0:1 */
	mypanel = destwall->one[IPROC];
	{
		register double t = destwall->all.min.x;
		mypanel.min.x -= t;
		mypanel.max.x -= t;
		t = 1.0 / (destwall->all.max.x - t);
		mypanel.min.x *= t;
		mypanel.max.x *= t;
		t = destwall->all.min.y;
		mypanel.min.y -= t;
		mypanel.max.y -= t;
		t = 1.0 / (destwall->all.max.y - t);
		mypanel.min.y *= t;
		mypanel.max.y *= t;
	}

	/* Does the image overlap this PE display? */
	if ((destarea.min.x > mypanel.max.x) ||
	    (destarea.max.x < mypanel.min.x) ||
	    (destarea.min.y > mypanel.max.y) ||
	    (destarea.max.y < mypanel.min.y)) {
		/* Nope; nothing to display */
		return;
	}

	/* Compute portion of display that is affected */
	if (destarea.min.x < mypanel.min.x) {
		scale = ((mypanel.min.x - destarea.min.x) /
			 (destarea.max.x - destarea.min.x));
		srcarea.min.x += (scale * (srcarea.max.x - srcarea.min.x));
		destarea.min.x = mypanel.min.x;
	}
	if (destarea.max.x > mypanel.max.x) {
		scale = ((destarea.max.x - mypanel.max.x) /
			 (destarea.max.x - destarea.min.x));
		srcarea.max.x -= scale * (srcarea.max.x - srcarea.min.x);
		destarea.max.x = mypanel.max.x;
	}
	if (destarea.min.y < mypanel.min.y) {
		scale = ((mypanel.min.y - destarea.min.y) /
			 (destarea.max.y - destarea.min.y));
		srcarea.min.y += (scale * (srcarea.max.y - srcarea.min.y));
		destarea.min.y = mypanel.min.y;
	}
	if (destarea.max.y > mypanel.max.y) {
		scale = ((destarea.max.y - mypanel.max.y) /
			 (destarea.max.y - destarea.min.y));
		srcarea.max.y -= scale * (srcarea.max.y - srcarea.min.y);
		destarea.max.y = mypanel.max.y;
	}
	scale = (mypanel.max.x - mypanel.min.x);
	d.min.x = (vwpanel.dim.x * (destarea.min.x - mypanel.min.x)) / scale;
	d.max.x = (vwpanel.dim.x * (destarea.max.x - mypanel.min.x)) / scale;
	scale = (mypanel.max.y - mypanel.min.y);
	d.min.y = (vwpanel.dim.y * (destarea.min.y - mypanel.min.y)) / scale;
	d.max.y = (vwpanel.dim.y * (destarea.max.y - mypanel.min.y)) / scale;

	/* We know where it is going, now do it */
	vpres(&vwpanel,
	      d,
	      srcimage,
	      srcarea,
	      frac);
}
