/*	vwtimehelp.c

	Helper code to generate ticks per microsecond multiplier

	March 1999 by H. Dietz
*/


#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>
#include <sys/time.h>


typedef double vwtime;

static	struct timeval	sysref;

vwtime
vwnow(void)
{
	static struct timeval sysnow;
	register vwtime mynow;

	gettimeofday(&sysnow, 0);
	mynow = ((1000000.0 * (sysnow.tv_sec - sysref.tv_sec)) +
		 (sysnow.tv_usec - sysref.tv_usec));
	return(mynow * (1.0 / 1000000.0));
}

void
vwreftime(void)
{
	gettimeofday(&sysref, 0);
}

union tick_union {
	unsigned long long	val64;
	struct {
		unsigned int	low32;
		unsigned int	hi32;
	} lowhi;
};

static unsigned long long
vwnowTick(void)
{
	register union tick_union result;
	register unsigned int msr = 0;

	__asm __volatile(
".byte 0xf; .byte 0x31	# Read the cycle counter
	movl %%edx, %0	# high order bits
	movl %%eax, %1	# low order bits"
: "=g" (result.lowhi.hi32), "=g" (result.lowhi.low32)
: "g" (msr)
: "eax", "edx"
			 );

	return(result.val64);
}

/*	Calibration constants...
*/
#define	SAMPLES	8	/* Even number, >=4 */
#define	SLEEP	15	/* At least 10, longer is better */

int
main()
{
	register unsigned long long t0, t1, t0a, t1a;
	register vwtime r0, r1;
	register double tickspers, factor;
	double d[SAMPLES];
	register int i, j;

	fprintf(stderr,
		"Calibrating the processor clock will take %1.1f minutes\n",
		((SAMPLES * SLEEP) / 60.0));

	printf("/*\tvwtime.h\n\n\tMachine-dependent time constants\n*/\n\n");

	vwreftime();
	for (i=0; i<SAMPLES; ++i) {
		t0 = vwnowTick();
		r0 = vwnow();
		t0a = vwnowTick();
		usleep(SLEEP * 1000000);
		t1 = vwnowTick();
		r1 = vwnow();
		t1a = vwnowTick();
		t0 = ((t0a + t0) / 2);
		t1 = ((t1a + t1) / 2);
		d[i] = ((double) (t1 - t0)) / (r1 - r0);
	}

	/* Sort values */
	for (i=0; i<SAMPLES; ++i) {
		for (j=i+1; j<SAMPLES; ++j) {
			if (d[i] > d[j]) {
				register double t = d[i];
				d[i] = d[j];
				d[j] = t;
			}
		}
	}

	/* Median filter */
	printf("/*\tEmpirical Intel processor clock ticks/second\n");
	printf("\tApproximate error: %f%%\n*/\n",
	       ((100.0 * (d[SAMPLES / 2 + 1] - d[SAMPLES / 2 - 2])) /
		d[SAMPLES / 2 - 2]));
	tickspers = ((d[SAMPLES / 2 - 1] + d[SAMPLES / 2]) / 2.0);
	printf("#define VWINTELTICKS %10.1f\n\n", tickspers);

	/* Determine usleep(1) time */
	t0 = 0;
	for (i=0; i<SAMPLES; ++i) {
		t0a = vwnowTick();
		usleep(1);
		t1a = vwnowTick();
		t1a -= t0a;
		if (t1a > t0) t0 = t1a;
	}
	printf("/*\tEmpirical worst-case seconds for usleep(1)\n*/\n");
	printf("#define VWUSLEEP1TIME %1.10f\n\n", t0 / tickspers);

	/* Determine vwnow() time */
	t0 = 0;
	factor = (1.0 / tickspers);
	for (i=0; i<SAMPLES; ++i) {
		t0a = vwnowTick();
		t1 = (vwnowTick() * factor);
		t1a = vwnowTick();
		t1a -= t0a;
		if (t1a > t0) t0 = t1a;
	}
	printf("/*\tEmpirical worst-case seconds for vwnow()\n*/\n");
	printf("#define VWNOWTIME %1.10f\n", t0 / tickspers);

	return(0);
}
