/*	com.c

	AFAPI communication function definitions for WAPERS.

	July 1997 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"

/*	p_bcastPut/Get() for 1u and bar...

	These do not do any compression.
*/

inline p_1u
p_bcastPut1u(register p_1u datum)
{
	return(p__all(datum));
}


inline p_1u
p_bcastGet1u(void)
{
	return(p__all(1));
}


p_bar
p_bcastPutbar(register p_bar datum)
{
	register int bits = (NPROC - 1);

	do {
		p__all(datum & (1 << bits));
	} while (--bits >= 0);
	return(datum);
}

p_bar
p_bcastGetbar(void)
{
	register int bits = (NPROC - 1);
	register p_bar t = 0;

	do {
		t = ((t << 1) | p__all(1));
	} while (--bits >= 0);

	return(t);
}

/*	Object p_bcastPut/Get() using compression

	Each of these routines uses a single-element cache of the
	previous value, which is xored with the datum to be sent.  The
	8u, 8, 16u, 16, 32f, and 64f routines simply use the 0-nybble
	compression built-into p__get() and p__put().  Even for float
	values, this should have some effect because it is common that
	the exponent or even some mantissa bits will be identical from
	one value to the next.  The other routines do more, first
	normalizing the xor result to have a 0 most significant bit by
	inverting it if needed, then sending the run length of the
	leading zero sequence, then the data bits, and finally sending
	a bit that says if the result should be inverted.  All these
	schemes have in common that they aggressively zero-compress,
	and use xor with a single-element cached value to increase the
	probability of zeros.
*/
	
static	p_8u	p_bcastCache8u = 0;

p_8u
p_bcastPut8u(register p_8u datum)
{
	/* Exclusive or with previous byte */
	register p_8u xdatum = (datum ^ p_bcastCache8u);

	/* Send the (zero compressed) nybbles */
	p__put(xdatum >> 4);
	p__put(xdatum);

	/* Update the cache */
	return(p_bcastCache8u = datum);
}

p_8u
p_bcastGet8u(void)
{
	register p_8u datum = p__get();
	datum = ((p__get() | (datum << 4)) ^ p_bcastCache8u);
	return(p_bcastCache8u = datum);
}


static	p_8	p_bcastCache8 = 0;

p_8
p_bcastPut8(register p_8 datum)
{
	/* Exclusive or with previous byte */
	register p_8 xdatum = (datum ^ p_bcastCache8);

	/* Send the (zero compressed) nybbles */
	p__put(xdatum >> 4);
	p__put(xdatum);

	/* Update the cache */
	return(p_bcastCache8 = datum);
}

p_8
p_bcastGet8(void)
{
	register p_8 datum = p__get();
	datum = ((p__get() | (datum << 4)) ^ p_bcastCache8);
	return(p_bcastCache8 = datum);
}


static	p_16u	p_bcastCache16u = 0;

p_16u
p_bcastPut16u(register p_16u datum)
{
	register int waspos = 1;
	register int toppos = p_16u_bits;
	register p_16u xdatum = (p_bcastCache16u ^ datum);
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (1 << (p_16u_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_16u_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum & (((1 << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			p__all(xdatum & (1 << toppos));
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_16u_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	return(p_bcastCache16u = datum);
}

p_16u
p_bcastGet16u(void)
{
	register int toppos = p_16u_bits;
	register p_16u result = 0;
	register int runlen;

	for (runlen=(p_16u_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_16u_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCache16u);

	/* Update cache */
	return(p_bcastCache16u = result);
}


static	p_16	p_bcastCache16 = 0;

p_16
p_bcastPut16(register p_16 datum)
{
	register int waspos = 1;
	register int toppos = p_16_bits;
	register p_16 xdatum = (p_bcastCache16 ^ datum);
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (1 << (p_16_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_16_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum & (((1 << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			p__all(xdatum & (1 << toppos));
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_16_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	return(p_bcastCache16 = datum);
}

p_16
p_bcastGet16(void)
{
	register int toppos = p_16_bits;
	register p_16 result = 0;
	register int runlen;

	for (runlen=(p_16_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_16_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCache16);

	/* Update cache */
	return(p_bcastCache16 = result);
}


static	p_32u	p_bcastCache32u = 0;

p_32u
p_bcastPut32u(register p_32u datum)
{
	register int waspos = 1;
	register int toppos = p_32u_bits;
	register p_32u xdatum = (p_bcastCache32u ^ datum);
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (1 << (p_32u_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_32u_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum & (((1 << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			p__all(xdatum & (1 << toppos));
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_32u_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	return(p_bcastCache32u = datum);
}


p_32u
p_bcastGet32u(void)
{
	register int toppos = p_32u_bits;
	register p_32u result = 0;
	register int runlen;

	for (runlen=(p_32u_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_32u_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCache32u);

	/* Update cache */
	return(p_bcastCache32u = result);
}


static	p_32	p_bcastCache32 = 0;

p_32
p_bcastPut32(register p_32 datum)
{
	register int waspos = 1;
	register int toppos = p_32_bits;
	register p_32 xdatum = (p_bcastCache32 ^ datum);
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (1 << (p_32_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_32_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum & (((1 << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			p__all(xdatum & (1 << toppos));
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_32_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	return(p_bcastCache32 = datum);
}


p_32
p_bcastGet32(void)
{
	register int toppos = p_32_bits;
	register p_32 result = 0;
	register int runlen;

	for (runlen=(p_32_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_32_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCache32);

	/* Update cache */
	return(p_bcastCache32 = result);
}


static	p_64u	p_bcastCache64u = 0;

p_64u
p_bcastPut64u(register p_64u datum)
{
	register int waspos = 1;
	register int toppos = p_64u_bits;
	register p_64u xdatum = (p_bcastCache64u ^ datum);
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (((p_64u) 1) << (p_64u_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_64u_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum &
			    (((((p_64u) 1) << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			register p_1u t = (((p_64u) 0) !=
					   (xdatum & (((p_64u) 1) << toppos)));
			p__all(t);
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_64u_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	return(p_bcastCache64u = datum);
}


p_64u
p_bcastGet64u(void)
{
	register int toppos = p_64u_bits;
	register p_64u result = 0;
	register int runlen;

	for (runlen=(p_64u_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_64u_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCache64u);

	/* Update cache */
	return(p_bcastCache64u = result);
}


static	p_64	p_bcastCache64 = 0;

p_64
p_bcastPut64(register p_64 datum)
{
	register int waspos = 1;
	register int toppos = p_64_bits;
	register p_64 xdatum = (p_bcastCache64 ^ datum);
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (((p_64) 1) << (p_64_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_64_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum &
			    (((((p_64) 1) << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			register p_1u t = (((p_64) 0) !=
					   (xdatum & (((p_64) 1) << toppos)));
			p__all(t);
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_64_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	return(p_bcastCache64 = datum);
}


p_64
p_bcastGet64(void)
{
	register int toppos = p_64_bits;
	register p_64 result = 0;
	register int runlen;

	for (runlen=(p_64_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_64_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCache64);

	/* Update cache */
	return(p_bcastCache64 = result);
}


static	p_32u	p_bcastCacheptr = 0;

p_ptr
p_bcastPutptr(register p_ptr datum)
{
	register int waspos = 1;
	register int toppos = p_ptr_bits;
	register p_32u xdatum = (p_bcastCacheptr ^ ((p_32u) datum));
	register int runlen;

	/* Normalize to a leading zero */
	if (xdatum & (1 << (p_32u_bits - 1))) {
		xdatum = ~xdatum;
		waspos = 0;
	}

	if (xdatum) {
		/* Normal case is non-zero value */

		for (runlen=(p_32u_bits/2); runlen>0; runlen>>=1) {
			register int newtop = (toppos - runlen);

			if (xdatum & (((1 << runlen) - 1) << newtop)) {
				/* NOT runlen (more) 0-value MSBs */
				p__all(0);
			} else {
				/* There were; decrement toppos */
				p__all(1);
				toppos = newtop;
			}
		}

		while (--toppos >= 0) {
			p__all(xdatum & (1 << toppos));
		}
	} else {
		/* If it is all zero, send all 0 */
		for (runlen=(p_32u_bits/2); runlen>0; runlen>>=1) {
			p__all(0);
		}
	}

	/* Let the other guys know if normalization inverted it */
	p__all(waspos);

	/* Update cache */
	p_bcastCacheptr = ((p_32u) datum);
	return(datum);
}


p_ptr
p_bcastGetptr(void)
{
	register int toppos = p_32u_bits;
	register p_32u result = 0;
	register int runlen;

	for (runlen=(p_32u_bits/2); runlen>0; runlen>>=1) {
		if (p__all(1)) {
			/* Adjust for run of 0-value MSBs */
			toppos -= runlen;
		}
	}

	if (toppos != p_32u_bits) {
		/* Wasn't all 0; get the rest of the bits */
		while (--toppos >= 0) {
			result += (result + p__all(1));
		}
	}

	/* Correct for normalization and xor with cache */
	result = ((p__all(1) ? result : ~result) ^ p_bcastCacheptr);

	/* Update cache */
	return((p_ptr) (p_bcastCacheptr = result));
}


static	p_32u	p_bcastCache32f = 0;	/* Note: cache is an int */

p_32f
p_bcastPut32f(register p_32f datum)
{
	register p_32u xdatum = p_bcastCache32f;
	register int i;

	/* Convert float value into int bit pattern */
	*((p_32f *) &p_bcastCache32f) = datum;
	xdatum ^= p_bcastCache32f;

	/* No compression beyond nybbles */
	for (i=0; i<p_32f_bits; i+=4) {
		p__put(xdatum & 0xf);
		xdatum >>= 4;
	}
	return(datum);
}

p_32f
p_bcastGet32f(void)
{
	register p_32u xdatum = 0;
	register int i;

	/* No compression beyond nybbles */
	for (i=0; i<p_32f_bits; i+=4) {
		xdatum |= (p__get() << i);
	}
	p_bcastCache32f ^= xdatum;

	/* Return int bit pattern as float value */
	return(*((p_32f *) &p_bcastCache32f));
}


static	p_64u	p_bcastCache64f = 0;	/* Note: cache is an int */

p_64f
p_bcastPut64f(register p_64f datum)
{
	register p_64u xdatum = p_bcastCache64f;
	register int i;

	/* Convert float value into int bit pattern */
	*((p_64f *) &p_bcastCache64f) = datum;
	xdatum ^= p_bcastCache64f;

	/* No compression beyond nybbles */
	for (i=0; i<p_64f_bits; i+=4) {
		p__put(((int) xdatum) & 0xf);
		xdatum >>= 4;
	}
	return(datum);
}

p_64f
p_bcastGet64f(void)
{
	register p_64u xdatum = 0;
	register int i;

	/* No compression beyond nybbles */
	for (i=0; i<p_64f_bits; i+=4) {
		xdatum |= (((p_64u) p__get()) << i);
	}
	p_bcastCache64f ^= xdatum;

	/* Return int bit pattern as float value */
	return(*((p_64f *) &p_bcastCache64f));
}

/*	p_bcastPutz() and p_bcastGetz()...

	We have already developed a very fancy hash-indexed, cached,
	run-length encoded, compression scheme...  but this does not
	use it.  We will distribute that version when we have more
	information about the type of data these routines are used to
	transmit, so that we can better tune the hashing, etc.

	For now, these routines simply use the byte routines... heck,
	they already can yield up to 4:1 compression.
*/

inline void
p_bcastPutz(register void *source,
register int bytes)
{
	while (bytes-- > 0) {
		p_bcastPut8u(*((p_8u *) source));
		++source;
	}
}


inline void
p_bcastGetz(register void *dest,
register int bytes)
{
	if (dest) {
		while (bytes-- > 0) {
			*((p_8u *) dest) = p_bcastGet8u();
			++dest;
		}
	} else {
		while (bytes-- > 0) {
			p_bcastGet8u();
		}
	}
}

/*	p_putGet() functions...

	The basic algorithm is to perform a sequence of broadcasts,
	however, this sequence skips broadcasts from PEs whose values
	will not be used by any processor.
*/

p_1u
p_putGet1u(register p_1u datum,
register int source)
{
	return((p__waitbar(datum) & (1 << source)) ? 1 : 0);
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_putGet##suffix(register p_##suffix datum, \
register int source) \
{ \
	register p_##suffix d = datum; \
	register int who = 0; \
 \
	if (source == IPROC) source = NPROC; \
	do { \
		if (p_any(source == who)) { \
			if (IPROC == who) { \
				/* Our turn to send and store data */ \
				p_bcastPut##suffix(datum); \
			} else if (source == who) { \
				/* We just get and store data... */ \
				d = p_bcastGet##suffix(); \
			} else { \
				/* Just be there for whoever... */ \
				p_bcastGet##suffix(); \
			} \
		} \
	} while (++who < NPROC); \
 \
	return(d); \
}

P_DECLARE(bar)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)
P_DECLARE(ptr)

#undef	P_DECLARE

void
p_putGetz(register void *dest,
register void *src,
register int n,
register int source)
{
	register int who = 0;
	register p_bar mask =
		p_reduceOrbar(((source == IPROC) ||
			       (dest == ((void *) 0))) ?
			       0 :
			       (1 << source));

	do {
		if (mask & 1) {
			if (IPROC == who) {
				/* Our turn to send data */
				p_bcastPutz(src, n);
			} else if (source == who) {
				/* We just get and store data... */
				p_bcastGetz(dest, n);
			} else {
				/* Just be there for whoever... */
				p_bcastGetz(((void *) 0), n);
			}
		}
		++who;
		mask >>= 1;
	} while (who < NPROC);

	if ((source == IPROC) && dest) {
		memcpy(dest, src, n);
	}
}

/*	p_gather() functions...
*/
#define	P_DECLARE(suffix) \
inline void \
p_gather##suffix(register p_##suffix *dest, \
register p_##suffix datum) \
{ \
	register p_##suffix d; \
	register int who = 0; \
 \
	do { \
		if (IPROC == who) { \
			/* Our turn to send and store data */ \
			p_bcastPut##suffix(d = datum); \
		} else { \
			/* We just get and store data... */ \
			d = p_bcastGet##suffix(); \
		} \
 \
		if (dest) dest[who] = d; \
	} while (++who < NPROC); \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_gatherz(register void *dest,
register void *src,
register int n)
{
	register int who = 0;

	do {
		if (IPROC == who) {
			/* Our turn to send and store data */
			p_bcastPutz(src, n);
			if (dest) {
				memcpy(dest, src, n);
			}
		} else {
			/* We just get and store data... */
			p_bcastGetz(dest, n);
		}

		if (dest) dest += n;
	} while (++who < NPROC);
}


/*	p_scatPut() and p_scatGet() functions...

	This is really a very simple operation to construct using the
	other routines, but it has been added to the library by
	popular demand....
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scatPut##suffix(register p_##suffix *src) \
{ \
	register int who = 0; \
 \
	do { \
		p_bcastPut##suffix(src[who]); \
	} while (++who < NPROC); \
 \
	return(src[IPROC]); \
} \
 \
p_##suffix \
p_scatGet##suffix(void) \
{ \
	register p_##suffix t; \
	register p_##suffix result = 0; \
	register int who = 0; \
 \
	do { \
		t = p_bcastGet##suffix(); \
		if (who == IPROC) result = t; \
	} while (++who < NPROC); \
 \
	return(result); \
}

P_ALPHA

#undef	P_DECLARE

void
p_scatPutz(register void *dest,
register void *src,
register int n)
{
	register int who = 0;

	do {
		p_bcastPutz(src, n);
		if (dest && (who == IPROC)) {
			memcpy(dest, src, n);
		}
		src += n;
	} while (++who < NPROC);
}

void
p_scatGetz(register void *dest,
register int n)
{
	register int who = 0;

	do {
		p_bcastGetz(((who == IPROC) ? dest : ((void *) 0)), n);
	} while (++who < NPROC);
}

/*	p_all2all() functions...

	The meaning of all2all here is essentially the same as that
	used in MPI, namely, every PE has different data for each PE.
	This is more properly called "personalized all-to-all," but so
	much for terminology....  If our hardware supports putGet,
	then NPROC-1 putGet should be used; WAPERS does not support
	putGet directly, but it is close enough....
*/
#define	P_DECLARE(suffix) \
inline void \
p_all2all##suffix(register p_##suffix *dest, \
register p_##suffix *src) \
{ \
	register int dist = 1; \
 \
	if (dest) { \
		dest[IPROC] = src[IPROC]; \
		do { \
			register int to = (IPROC + dist); \
			register int from = (IPROC - dist); \
 \
			if (to >= NPROC) to -= NPROC; \
			if (from < 0) from += NPROC; \
			dest[from] = p_putGet##suffix(src[to], from); \
		} while (++dist < NPROC); \
	} else { \
		do { \
			register int to = (IPROC + dist); \
 \
			if (to >= NPROC) to -= NPROC; \
			p_putGet##suffix(src[to], NPROC); \
		} while (++dist < NPROC); \
	} \
}

P_ALPHA

#undef	P_DECLARE

inline void
p_all2allz(register void *dest,
register void *src,
register int n)
{
	register int dist = 1;

	if (dest) {
		register int off = (n * IPROC);

		memcpy(dest + off, src + off, n);
		do {
			register int to = (IPROC + dist);
			register int from = (IPROC - dist);

			if (to >= NPROC) to -= NPROC;
			if (from < 0) from += NPROC;
			p_putGetz(dest + (from * n),
				  src + (to * n),
				  n,
				  from);
		} while (++dist < NPROC);
	} else {
		do {
			register int to = (IPROC + dist);

			if (to >= NPROC) to -= NPROC;
			p_putGetz(((void *) 0),
				  src + (to *n),
				  n,
				  IPROC);
		} while (++dist < NPROC);
	}
}
#endif /* if NPROC > 1 */
