/*	intern.h

	AFAPI internals for WAPERS.

	September 1998 by H. Dietz and T.I.Mattox
*/
#if	NPROC == 1
#include "../ONE/one.h"
#endif
#if	NPROC > 1


#ifndef _P_INTERN_H
#define _P_INTERN_H

#include <AFAPI/afapi.h>

/*	Hardware port stuff...
*/
extern unsigned short p_modeport;
extern unsigned char p_outprev;
extern unsigned int p_outbit;


/*
	AFAPI WAPERS inline primitive functions....

	WAPERS is literally nothing more than a 4-bit wide wired-AND
	of the open-collector SPP input/output bits:  pins 1, 14, 16,
	and 17 each should be wired-AND across all PEs, and pins 18-25
	should form the common signal ground for all PEs.  That's it.

	How does that do anything useful?  Well....

	All magic constants have been written out explicitly to
	minimize potential name conflicts with user code.  Since it no
	longer damages readability ;-), we also have made a few
	optimizations which are dependent on the particular bit
	assignments within the port registers.  The other major change
	is that port addresses are now variables set by p_init(), thus
	allowing a single object code to run on machines with parallel
	ports at various different addresses.

	July 1997 by Hank Dietz
*/



/*	ioperm() and iopl()

	Function prototypes included here since it's location in the standard
	header file tree has moved from version to version of Linux:
		Under Libc5 these are in <unistd.h>
		Under Glibc1 these are left unspecified!
		Under Glibc2 these are in <sys/io.h>  (and <sys/perm,h>)
*/
extern int
ioperm(unsigned long from, unsigned long num, int turn_on);

#ifdef P_USE_IOPL
extern int
iopl(int level);
#endif


/*	p__inb() and p__outb()...

	These are the actual lowest-level hardware port I/O
	instructions used to touch the parallel port interface.  They
	are defined here, rather than borrowed from a library, to
	ensure that they don't add any delays...  many library
	versions insert code to allow port settle time, which we
	confirm using a handshake protocol instead.
*/

#ifdef	P_DEBOUNCE

extern inline
unsigned int
p___inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

extern inline
unsigned int
p__inb(unsigned short port)
{
	register unsigned int t = p___inb(port);

	while (t != p_inlast) {
		p_inlast = t;
		t = p___inb(port);
	}
	return(t);
}

#else

extern inline
unsigned int
p__inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

#endif


extern inline
void
p__outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}

/*	p__wait()...

	The basic logic is pretty simple, and we get a two-cycle
	barrier, but the interesting fact is that with only wired-AND,
	it takes using two separate bits of three reserved for
	barriers to implement each barrier.

	The strange little xors compensate for the fact that three of
	the four open-collector output bits are inverted.
*/

extern inline
void
p__wait(void)
{
	register int p_outnext =
		((p_outprev == 0x2) ? 0x8 : (p_outprev >> 1));

	/* Set previous and next barrier bits */
	p__outb(((p_outprev | p_outnext | p_outbit) ^
		 (0x8 | 0x2 | 0x1)), p_modeport);

	/* Wait for wired-AND of next bit to be 1 */
	while (((p__inb(p_modeport) ^
		 (0x8 | 0x2 | 0x1)) &
		p_outnext) == 0) ;

	/* Set previous bit for next time */
	p_outprev = p_outnext;
}

/*	p__all() and p__any()

	These do all the data transmission work....  Actually, it is
	all done by p__all(), since this perfectly matches the one-bit
	wired-AND data path of WAPERS.
*/

extern inline
p_1u
p__all(register p_1u f)
{
	p__wait();
	p_outbit = (f != 0);
	p__wait();
	return((p__inb(p_modeport) ^ (0x8 | 0x2 | 0x1)) & 0x1);
}

extern inline
p_1u
p__any(register p_1u f)
{
	return(!p__all(!f));
}

/*	p__nand()...

	A 4 bit wide NAND accomplished with four p__all calls...
*/	

extern inline
int
p__nand(register p_8u b)
{
	register int t;

	/* Compute 4-bit AND, then invert it */
	b &= 0x0f;
	t = p__all(b & 0x08);
	t = ((t << 1) | p__all(b & 0x04));
	t = ((t << 1) | p__all(b & 0x02));
	return(0xf ^ ((t << 1) | p__all(b & 0x01)));
}

/*	p__put() and p__get()...

	This takes advantage of the fact that the code
	doesn't have to be the same for sender and reciever.
*/	

extern inline
int
p__get(void)
{
	/* Get a nybble, with zero compression */

	if (p__all(1)) {
		register int t = p__all(1);
		t = ((t << 1) | p__all(1));
		t = ((t << 1) | p__all(1));
		return((t << 1) | p__all(1));
	} else {
		return(0);
	}
}


extern inline
void
p__put(register p_8u b)
{
	/* Put a nybble, with zero compression */

	b &= 0x0f;
	if (b) {
		p__all(1);
		p__all(b & 0x08);
		p__all(b & 0x04);
		p__all(b & 0x02);
		p__all(b & 0x01);
	} else {
		p__all(0);
	}
}

/*	p__waitbar()

	This really is not a bottom-level primitive for the current
	WAPERS hardware, but is small enough to inline....
*/

extern inline
p_bar
p__waitbar(register p_1u f)
{
	/* Collect a barrier bit mask with the votes from all PEs. Loop
	   collecting highest bits first...  very straightforward.
	*/
	register p_bar t = 0;
	register int i = (NPROC - 1);

	do {
		t = ((t << 1) | p__all((i != IPROC) || f));
	} while (--i >= 0);

	return(t);
}

/*	Inline signal primitives...

	Very simple stuff...  barrier NAND hardware does signals very
	naturally.
*/

extern inline void
s__signal(register int s)
{
	/* This is not worth inlining... */
	s_signal(s);
}


extern inline p_1u
s__pending(void)
{
	/* Really easy when you got the right hardware ;-) */
	return(((p__inb(p_modeport) ^ (0x8 | 0x2 | 0x1)) & 1) == 0);
}


extern inline void
s__poll(void)
{
	/* Really easy when you got the right hardware ;-) */
	if (((p__inb(p_modeport) ^ (0x8 | 0x2 | 0x1)) & 1) == 0) {
		s_signal(S_SIGANY);
	}
}


extern inline void
s__wait(void)
{
	/* Just exit and re-enter... that flushes things */
	s_exit();
	s_init();
}

#endif
#endif /* if NPROC > 1 */
