/*	reduce.c

	AFAPI associative reduction function definitions for WAPERS.

	July 1997 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"

/*	p_reduceAnd() functions...
*/

p_1u
p_reduceAnd1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceAnd##suffix(register p_##suffix datum) \
{ \
	register int bits = p_##suffix##_bits; \
	register p_##suffix result = 0; \
 \
	do { \
		register p_1u t = (((p_##suffix) 0) != \
				   (datum & (((p_##suffix) 1) << --bits))); \
		result += (result + p__all(t)); \
	} while (bits > 0); \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE

p_ptr
p_reduceAndptr(register p_ptr datum)
{
	return((p_ptr) p_reduceAnd32u((p_32u) datum));
}

/*	p_reduceOr() functions...
*/

inline p_1u
p_reduceOr1u(register p_1u datum)
{
	return(p__any(datum));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceOr##suffix(register p_##suffix datum) \
{ \
	register int bits = p_##suffix##_bits; \
	register p_##suffix result = 0; \
 \
	do { \
		register p_1u t = (((p_##suffix) 0) != \
				   (datum & (((p_##suffix) 1) << --bits))); \
		result += (result + p__any(t)); \
	} while (bits > 0); \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE

p_ptr
p_reduceOrptr(register p_ptr datum)
{
	return((p_ptr) p_reduceOr32u((p_32u) datum));
}

/*	p_reduceXor() functions...

	Xor is kinda interesting because it is bounded by Or, which is
	to say that only 1 bits in the Or could be 1 bits in the Xor.
	Thus, we can compute the Or and then vote on potential 1
	bits...  this does not always result in a faster operation,
	but if not, the additional cost is minimal.
*/

p_1u
p_reduceXor1u(register p_1u datum)
{
	return(p_count(datum) & 1);
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceXor##suffix(register p_##suffix datum) \
{ \
	register int bits = p_##suffix##_bits; \
	register p_##suffix result = 0; \
 \
	do { \
		register p_1u t = (((p_##suffix) 0) != \
				   (datum & (((p_##suffix) 1) << --bits))); \
		result += (result + p_reduceXor1u(t)); \
	} while (bits > 0); \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE

p_ptr
p_reduceXorptr(register p_ptr datum)
{
	return((p_ptr) p_reduceXor32u((p_32u) datum));
}

/*	p_reduceMin() functions...

	Most of these functions use an interesting systolic voting
	algorithm that actually makes min the same cost as broadcast
	(ignoring potential compression on broadcast), no matter how
	many processors are participating.  Signed integers are
	converted to unsigned by adding a bias factor, then the
	minimum is converted back.  We also take advantage of the fact
	that using integer compares to compare floating point values
	works for many floating point formats, thus the integer voting
	algorithm still works.
*/

p_1u
p_reduceMin1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	register int bitpos = (p_##suffix##_bits - 1); \
	register p_##suffix result = 0; \
 \
	do { \
		register p_##suffix t = datum & (((p_##suffix) 1) << bitpos); \
		result <<= 1; \
		if (p__all(((p_1u) (t != ((p_##suffix) 0))))) { \
			/* Min value has a 1 here */ \
			result |= 1; \
		} else { \
			/* Min value has a 0 here... \
			   if I do not, then I am not the min, \
			   and I lose all future votes \
			*/ \
			if (t) datum = p_##suffix##_ALL1; \
		} \
	} while (--bitpos >= 0); \
	return(result); \
}

P_DECLARE(8u)
P_DECLARE(16u)
P_DECLARE(32u)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE
#define	P_DECLARE(suffix, bias) \
p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	return(((p_##suffix) p_reduceMin##suffix##u(datum + bias)) - bias); \
}

P_DECLARE(8, (((p_32u) 1) << 7))
P_DECLARE(16, (((p_32u) 1) << 15))
P_DECLARE(32, (((p_64u) 1) << 31))
P_DECLARE(64, (((p_64u) 1) << 63))

#undef	P_DECLARE

#ifndef	P_USE_FLOAT_COMPARE

p_32f
p_reduceMin32f(register p_32f datum)
{
	p_32f result = datum;
	*((p_32 *) &result) = p_reduceMin32(*((p_32 *) &result));
	return(result);
}

p_64f
p_reduceMin64f(register p_64f datum)
{
	p_64f result = datum;
	*((p_64 *) &result) = p_reduceMin64(*((p_64 *) &result));
	return(result);
}

#else	/* if P_USE_FLOAT_COMPARE */

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMin##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix min = datum; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
		} else { \
			register p_##suffix t = p_bcastGet##suffix(); \
			if (t < min) min = t; \
		} \
	} while (who > 0); \
 \
	return(min); \
}

P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

#endif


p_ptr
p_reduceMinptr(register p_ptr datum)
{
	return((p_ptr) p_reduceMin32u((p_32u) datum));
}

/*	p_reduceMax() functions...

	This is essentially the same as the p_reduceMin() stuff.
*/

p_1u
p_reduceMax1u(register p_1u datum)
{
	return(p__any(datum));
}


#define	P_DECLARE(suffix) \
inline p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	register int bitpos = (p_##suffix##_bits - 1); \
	register p_##suffix result = 0; \
 \
	do { \
		register p_##suffix t = datum & (((p_##suffix) 1) << bitpos); \
		result <<= 1; \
		if (p__any(((p_1u) (t != ((p_##suffix) 0))))) { \
			/* Max value has a 1 here... \
			   if I do not, then I am not the max, \
			   and I lose all future votes \
			*/ \
			result |= 1; \
			if (!t) datum = 0; \
		} \
	} while (--bitpos >= 0); \
	return(result); \
}

P_DECLARE(8u)
P_DECLARE(16u)
P_DECLARE(32u)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE
#define	P_DECLARE(suffix, bias) \
p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	return(((p_##suffix) p_reduceMax##suffix##u(datum + bias)) - bias); \
}

P_DECLARE(8, (((p_32u) 1) << 7))
P_DECLARE(16, (((p_32u) 1) << 15))
P_DECLARE(32, (((p_64u) 1) << 31))
P_DECLARE(64, (((p_64u) 1) << 63))

#undef	P_DECLARE

#ifndef	P_USE_FLOAT_COMPARE

p_32f
p_reduceMax32f(register p_32f datum)
{
	p_32f result = datum;
	*((p_32 *) &result) = p_reduceMax32(*((p_32 *) &result));
	return(result);
}

p_64f
p_reduceMax64f(register p_64f datum)
{
	p_64f result = datum;
	*((p_64 *) &result) = p_reduceMax64(*((p_64 *) &result));
	return(result);
}

#else	/* if P_USE_FLOAT_COMPARE */

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMax##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix max = datum; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
		} else { \
			register p_##suffix t = p_bcastGet##suffix(); \
			if (t > max) max = t; \
		} \
	} while (who > 0); \
 \
	return(max); \
}

P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

#endif


p_ptr
p_reduceMaxptr(register p_ptr datum)
{
	return((p_ptr) p_reduceMax32u((p_32u) datum));
}

/*	p_reduceAdd() functions...

	Slow and dumb...  WAPERS just isn't good at addition.
*/

p_1u
p_reduceAdd1u(register p_1u datum)
{
	/* Result is just 0 or 1 */
	return(p__any(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceAdd##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix sum = 0; \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
			sum += datum; \
		} else { \
			sum += p_bcastGet##suffix(); \
		} \
	} while (who > 0); \
 \
	return(sum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_reduceMul() functions...

	Slow and dumb...  WAPERS certainly cannot multiply.  The only
	trick here is that we do know how to multiply by zero really
	quickly...  so if any PE has 0, we can skip the whole mess.
*/

p_1u
p_reduceMul1u(register p_1u datum)
{
	return(p__all(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_reduceMul##suffix(register p_##suffix datum) \
{ \
	register int who = NPROC; \
	register p_##suffix product = 1; \
 \
	if (p__any((p_1u) (datum == ((p_##suffix) 0)))) return(0); \
 \
	do { \
		if (--who == IPROC) { \
			p_bcastPut##suffix(datum); \
			product *= datum; \
		} else { \
			product *= p_bcastGet##suffix(); \
		} \
	} while (who > 0); \
 \
	return(product); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE
#endif /* if NPROC > 1 */
