/*	scan.c

	AFAPI scan (parallel prefix) function definitions for
	WAPERS.

	July 1997 by H. Dietz
*/
#if	NPROC > 1

#include "intern.h"

/*	p_scanAnd() functions...

	Sneaky implementation uses p_first for each bit position,
	detecting the lowest numbered (first) PE that has a 0 bit.
	Note: p_first() takes base 2 log(NPROC + 1) steps.
*/

inline p_1u
p_scanAnd1u(register p_1u datum)
{
	return(IPROC < p_first(datum == 0));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanAnd##suffix(register p_##suffix datum) \
{ \
	register int bits = p_##suffix##_bits; \
	register p_##suffix result = 0; \
 \
	do { \
		register p_1u t = (((p_##suffix) 0) != \
				   (datum & (((p_##suffix) 1) << bits))); \
		result = ((result << 1) | p_scanAnd1u(t)); \
	} while (--bits >= 0); \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE


p_ptr
p_scanAndptr(register p_ptr datum)
{
	return((p_ptr) p_scanAnd32u((p_32u) datum));
}

/*	p_scanOr() functions...

	Sneaky implementation uses p_first for each bit position,
	detecting the lowest numbered (first) PE that has a 1 bit.
	Note: p_first() takes base 2 log(NPROC + 1) steps.
*/

inline p_1u
p_scanOr1u(register p_1u datum)
{
	return(IPROC >= p_first(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanOr##suffix(register p_##suffix datum) \
{ \
	register int bits = p_##suffix##_bits; \
	register p_##suffix result = 0; \
 \
	do { \
		register p_1u t = (((p_##suffix) 0) != \
				   (datum & (((p_##suffix) 1) << bits))); \
		result = ((result << 1) | p_scanOr1u(t)); \
	} while (--bits >= 0); \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE


p_ptr
p_scanOrptr(register p_ptr datum)
{
	return((p_ptr) p_scanOr32u((p_32u) datum));
}

/*	p_scanXor() functions...

	Implementation does NPROC-1 full-value broadcasts.
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanXor##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix sum = 0; \
	register p_##suffix result = 0; \
 \
	do { \
		if (who == IPROC) { \
			result = (sum ^= datum); \
			p_bcastPut##suffix(sum); \
		} else { \
			sum = p_bcastGet##suffix(); \
		} \
	} while (++who < (NPROC - 1)); \
 \
	return((IPROC == (NPROC-1)) ? (sum ^ datum) : result); \
}

P_DECLARE(1u)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(bar)

#undef	P_DECLARE


p_ptr
p_scanXorptr(register p_ptr datum)
{
	return((p_ptr) p_scanXor32u((p_32u) datum));
}

/*	p_scanMin() functions...

	Sneaky implementation outputs the PE0 value and then votes to
	see if the PE would have a new minimum.  If it does, then that
	PE will output its value.  Otherwise, we continue the voting
	with the next PE.
*/

p_1u
p_scanMin1u(register p_1u datum)
{
	return(p_scanAnd1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMin##suffix(register p_##suffix datum) \
{ \
	register int who; \
	register p_##suffix min; \
	register p_##suffix result = 0; \
 \
	if (IPROC == 0) { \
		p_bcastPut##suffix(datum); \
		result = (min = datum); \
	} else { \
		min = p_bcastGet##suffix(); \
	} \
	for (who=1; who<NPROC; ++who) { \
		if (who == IPROC) { \
			if (datum < min) { \
				p_all(1); \
				p_bcastPut##suffix(datum); \
				min = datum; \
			} else { \
				p_all(0); \
			} \
			result = min; \
		} else { \
			if (p_all(1)) { \
				min = p_bcastGet##suffix(); \
			} \
		} \
	} \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanMax() functions...

	Done like p_scanMin().
*/

p_1u
p_scanMax1u(register p_1u datum)
{
	return(p_scanOr1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMax##suffix(register p_##suffix datum) \
{ \
	register int who; \
	register p_##suffix max; \
	register p_##suffix result = 0; \
 \
	if (IPROC == 0) { \
		p_bcastPut##suffix(datum); \
		result = (max = datum); \
	} else { \
		max = p_bcastGet##suffix(); \
	} \
	for (who=1; who<NPROC; ++who) { \
		if (who == IPROC) { \
			if (datum > max) { \
				p_all(1); \
				p_bcastPut##suffix(datum); \
				max = datum; \
			} else { \
				p_all(0); \
			} \
			result = max; \
		} else { \
			if (p_all(1)) { \
				max = p_bcastGet##suffix(); \
			} \
		} \
	} \
 \
	return(result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(ptr)
P_DECLARE(bar)
P_DECLARE(32f)
P_DECLARE(64f)

#undef	P_DECLARE

/*	p_scanAdd() functions...

	Do the obvious, N-1 transmission, thing.
*/

p_1u
p_scanAdd1u(register p_1u datum)
{
	return(p_scanOr1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanAdd##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix sum = 0; \
 \
	do { \
		if (who < IPROC) { \
			sum += p_bcastGet##suffix(); \
		} else if (who == IPROC) { \
			p_bcastPut##suffix(datum); \
			sum += datum; \
		} else { \
			p_bcastGet##suffix(); \
		} \
	} while (++who < (NPROC-1)); \
 \
	return((IPROC == (NPROC-1)) ? (sum + datum) : sum); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef P_DECLARE

/*	p_scanMul() functions...

	Like p_scanAdd(), do the obvious, N-1 transmission, thing.
	However, stop everything if product ever becomes 0, because
	future multiplies can't change that value.
*/

p_1u
p_scanMul1u(register p_1u datum)
{
	return(p_scanAnd1u(datum));
}

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMul##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_##suffix product = 1; \
	register p_##suffix result = 0; \
 \
	do { \
		if (who == IPROC) { \
			result = (product *= datum); \
			p_bcastPut##suffix(product); \
		} else { \
			product = p_bcastGet##suffix(); \
		} \
		if (product == 0) return(result); \
	} while (++who < (NPROC - 1)); \
 \
	return((IPROC == (NPROC - 1)) ? (product * datum) : result); \
}

P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)
P_DECLARE(32f)
P_DECLARE(64f)

#undef P_DECLARE
#endif /* if NPROC > 1 */
