/*	signal.c

	AFAPI signal definitions for WAPERS.

	There's only one data bit in WAPERS that we can use for
	asynchronous signals, but that is enough.  In fact, the code
	is very clean, rather like SHMAPERS AFAPI, because the nasty
	compression and such is all handled by the p_ routines.

	July 1997 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"


#define	S_SIGMAX	5	/* Really could be anything > 2 */
#define	S_EXIT		(S_SIGMAX - 1)


static int s_active = 0;	/* Signals enabled? */

/*	Pre-defined signal handlers...
*/

void
s_ignore(register p_bar m)
{
	/* Nothing to do...  that's why it's called s_ignore() */
}


static void
s_exitHandler(register p_bar m)
{
	/* Is every process making this request? */
	if (m == ((((p_bar) 1) << NPROC) - ((p_bar)1))) {
		/* Yup, everybody wants out...  do it */
		s_active = 0;
	}
}

/*	s_vector[] array...

	The following is the magic internal structure used to vector
	signals to the correct handlers.  It is specific to WAPERS
	(although SHMAPERS AFAPI has the identical structure).
	The entry for s_exitHandler must be in the S_SIGMAX-1 position.
*/

static void (*s_vector[S_SIGMAX])(register p_bar mask) = {
	/* Normal signals... */
	s_ignore,
	s_ignore,
	s_ignore,
	s_ignore,

	/* And one signal you can't redefine... */
	s_exitHandler
};

/*	s_install()

	Install a signal handler...
*/

int
s_install(register int s,
register void (*fn)(p_bar m))
{
	/* Install the signal handler fn()... */
	register int which = s;

	/* If s is S_SIGANY, pick one that is s_ignore()... */
	if (which == S_SIGANY) {
		for (which = 0; which<(S_SIGMAX-1); ++which) {
			if (s_vector[which] == s_ignore) break;
		}
	}

	/* Got a valid signal number? */
	if ((which < 0) || (which >= (S_SIGMAX-1))) {
		/* Nope.  Return with error. */
		return(-1);
	}

	/* Ok, install it... */
	s_vector[which] = fn;

	/* Return signal number */
	return(which);
}

/*	s_init() and s_exit()...
*/

int
s_init(void)
{
	/* If it's already active, oops... */
	if (s_active) return(1);

	/* Ok, make it active... */
	p_outbit = 1;
	p_wait();
	p_wait();
	s_active = 1;
	return(0);
}


void
s_exit(void)
{
	/* Shut-down the signal system using a special non-signal... */
	while (s_active) {
		s_signal(S_EXIT);
	}
}

/*	s_signal()...

	This code must recognize who has sent signals, create masks grouping
	similar signals together, and invoke the appropriate handler
	functions in a way that gives s_exit the lowest priority....
*/

void
s_signal(register int s)
{
	register int sig;

	/* If we got here, somebody wants to send a signal.  This is
	   done by simply turning-off the data bit and then doing a
	   barrier synchronization...
	   However, if the signal being sent is S_EXIT, we should not turn
	   off the data bit...If no other signals are being asserted, and
	   every PE has called s_exit(), they will all call s_signal(S_EXIT)
	   even though s_poll/s_pending would indicate no signal is there.
	   Thus s_exit succeeds without forcing a pile of communications
	   whenever anyone does an s_poll.
	*/
	p_outbit = (s == S_EXIT);
	p__wait();
	p__wait();

	/* Ok, we're in p_ land now, so let's just figure out who
	   wants to do what and do them.  Notice that checking if
	   everyone wants to exit happens naturally.
	*/
	for (sig=0; sig<S_SIGMAX; ++sig) {
		if (p__any(s == sig)) {
			/* Somebody wants this signal; do it for all
			   the requestors in one shot.
			*/
			s_vector[sig](p_waitbar(s == sig));
		}
	}

	/* Re-enable the signal system */
	p__wait();
	p_outbit = 1;
	p__wait();
	p__wait();
}

/*	Inlined signal stuff...

	These are just callable versions of the simple interface
	routines....
*/

p_1u
s_pending(void)
{
	return(s__pending());
}

void
s_poll(void)
{
	s__poll();
}

void
s_wait(void)
{
	s__wait();
}
#endif /* if NPROC > 1 */
