/*	afdiscon.c

	Force this PE to act like its cable to the TTL_PAPERS unit
	was disconnected.

	1997 by Tim Mattox
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <AFAPI/version.h>


unsigned short p_outport;	/* Parallel port output address */
unsigned short p_inport;	/* Parallel port input address */
unsigned short p_modeport;	/* Mode address (used only by OS) */
unsigned char p_outlast;	/* Last port output value */

/*	ioperm() and iopl()

	Function prototypes included here since it's location in the standard
	header file tree has moved from version to version of Linux:
		Under Libc5 these are in <unistd.h>
		Under Glibc1 these are left unspecified!
		Under Glibc2 these are in <sys/io.h>  (and <sys/perm,h>)
*/
extern int
ioperm(unsigned long from, unsigned long num, int turn_on);

#ifdef P_USE_IOPL
extern int
iopl(int level);
#endif


extern inline
unsigned int
p__inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}


extern inline
void
p__outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}


int
main(register int argc,
register char **argv)
{
	register char *p;

	/* Step 0:  Examine environment variables */
	p = getenv("AFAPI");
	if ((p == NULL) || strcmp(AFAPI, p)) {
		fprintf(stderr,
"AFAPI reset:  need %s hardware (set AFAPI environment variable)\n",
			AFAPI);
		exit(1);
	}

	p = getenv("AFAPI_VERSION");
	if (p == NULL) p = "not given";
	if (strcmp(AFAPI_VERSION, p)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_VERSION of code is %s, environment is %s\n",
			AFAPI_VERSION, p);
		exit(1);
	}

	p = getenv("AFAPI_PORTBASE");
	if (p != NULL) {
		/* Set the port address variables */
		p_outport = (unsigned short) strtol(p, NULL, 0);
		p_inport = p_outport + 1;

#ifndef	P_NON_STANDARD_PORT
		/* Is this one of the three standard parallel port
		   addresses?  if not, we shouldn't trust it....  Of
		   course, there is nothing inherently wrong with
		   having a parallel port at another address, so this
		   is code an informed user can change if appropriate.
		*/
		if ((p_outport != 0x278) &&
		    (p_outport != 0x378) &&
		    (p_outport != 0x3bc)) {
			fprintf(stderr,
"AFAPI reset:  AFAPI_PORTBASE=%d is not a standard parallel port\n",
				p_outport);
			exit(1);
		}
#endif /* P_NON_STANDARD_PORT */
	} else {
		fprintf(stderr,
"AFAPI reset:  AFAPI_PORTBASE (parallel port address) was not set\n");
		exit(1);
	}


	/* Step 1:  Get permission to access the port.

	   This can use any of:

		ioperm(p_outport, 3, 1)
		ioperm(p_outport, 2, 1)
		iopl(3)

	   The ioperm() call is safer, but yields slower port
	   operations due to I/O protection checks and is
	   Linux-specific.  Because the mode port shouldn't be
	   accessed by user code, only 2, not 3, port registers need
	   to be accessible.
	*/
#ifdef P_USE_IOPL
        if (iopl(3))
#else
        if (ioperm(p_outport, 3, 1))
#endif
	{
		fprintf(stderr,
			"AFAPI PE:  can't access parallel port\n");
		exit(1);
	}
	p__outb(0x01, p_outport+2); /* set the mode port */

	p_outlast = (0xff);
	p__outb(p_outlast, p_outport);

	return(0);
}
