/*	ttltester.c

	This program is a torture test of the TTL_PAPERS 960801.
	It only tests the low level p__nand() routine... but if that's not
	working, there is no point in trying anything more complicated.

	1995 by Hank Dietz and Tim Mattox
	1996 converted to AFAPI by Tim Mattox
	1998 renamed ttltester and included in the distribution

	ver 980910
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <AFAPI/afapi.h>
#if NPROC > 1
#include <AFAPI/inline.h>


static long errs = 0;
static long billions = 0;
static long tests = 0;

void message(register int got, register int expect)
{
	fprintf(stderr,
"PE%02x test %ld%09ld, error %ld: got 0x%x (expected 0x%x, p_outlast = 0x%x)\n",
		IPROC, billions, tests, ++errs, got, expect, p_outlast);
	fprintf(stderr,"Press Control-C \n");
	fflush(stderr);

/* remove the next three lines to have it continue testing after an error */
	sleep(1000000);
	p_exit(); 
	exit(-1); 
}

void checksync(void)
{
	register int conf = p_confirm();

	if (conf != 0) {
		fprintf(stderr,
"PE%02x test %ld%09ld (%ld errors); p_confirm() returned %d!!!!!\n",
			IPROC, billions, tests, ++errs, conf);
		fflush(stderr);
	};
}

extern inline
void expect(register int got,
register int expect)
{
	if (++tests == 1000000000) {
		tests = 0;
		++billions;
	}
	if (got != expect) message(got, expect);
}


void flipper(register int times,
register int val1,
register int val2)
{
	/* Flips between val1 & val2 repeatedly "times". */
	register int i = times;

	while (i-- > 0) {
		expect(p__nand(val1), (val1 ^ 0xf));
		expect(p__nand(val2), (val2 ^ 0xf));
	}
}

void
tryflips(register int times)
{
	register int i;
	register int j;

	times /= (16 * 16 * 4);

	for (i=0; i<=0xf; ++i) {
		for (j=0; j<=0xf; ++j) {
			if (i != j) {
				flipper(times, i, j);
				/* Now flip the RDY and try again */
				p_wait();
				flipper(times, i, j);
			}
		}
	}
}

void
tryalone(register int times)
{
	register int pe;
	register int pat;
	register int i;

	times /= (NPROC * 16);

	for (pe=0; pe<NPROC; ++pe) {
		if (IPROC == CPROC) {
			fprintf(stderr,
"     test %ld%09ld (%ld errors); PE%d alone flipping phase...\n",
				billions, tests, errs, pe);
		}
		for (pat=0; pat<0xf; ++pat) {
			for (i=0; i<times; ++i) {
				expect(p__nand(0xf), 0);
				expect(p__nand((pe == IPROC) ? pat : 0xf),
					(pat ^ 0xf));
			}
		}
	}
}

#define	MAXPATS	(NPROC*NPROC)	/* Must be a power of 2 */

struct {
	char rtable[NPROC];
	char target;
} rpat[MAXPATS];

int
rand4()
{
	return((rand() >> 4) & 0xf);
}

void
randhelp()
{
	/* Build a table of random output values designed to
	   generate the randomly selected target value.
	   This is done to ensure that the output value will
	   generally be altered, whereas picking truly random
	   values would nearly always result in 0xf.
	*/
	register int j;
	register int i;
	register int k;

	for (k=0; k<MAXPATS; ++k) {
		j = 0xf;
		rpat[k].target = rand4();
		for (i=0; i<NPROC; ++i) {
			rpat[k].rtable[i] =
				((rpat[k].target ^ 0xf) | rand4());
			j &= rpat[k].rtable[i];
		}
		j ^= 0xf;

		rpat[k].rtable[rand() & (NPROC-1)] ^= (rpat[k].target ^ j);
	}
}
		
void
tryrand(register int times)
{
	register int i = 0;

	/* Get a set of randoms */
	randhelp();

	while (times-- > 0) {
		/* See if that worked... */
		expect(p__nand(rpat[i].rtable[iproc]), rpat[i].target);

		/* Move to next pattern */
		i = ((i + 1) & (MAXPATS - 1));
	}
}


int
main(register int argc,
register char **argv)
{
	register int batch_size, conf;

	if (argc != 2) {
		fprintf(stderr,
"Usage: %s <number_of_nands_per_test_phase>\n",
			argv[0]);
		exit(1);
	}

	batch_size = atoi(argv[1]);
	if (batch_size < 1) {
		fprintf(stderr,
"It's hard to have fewer than 1 nands per test phase!\n");
		exit(2);
	}

	/* check-in with PAPERS */
	if (p_init()) exit(3);

	while (1) {
		fprintf(stderr,
"PE%02x test %ld%09ld (%ld errors); one PE alone flipping phase...\n",
			IPROC, billions, tests, errs);
		tryalone(batch_size);
		fflush(stderr);

		p_wait();	/* Now flip the RDY and try again */
		fprintf(stderr,
"PE%02x test %ld%09ld (%ld errors); one PE alone flipping phase, RDY flipped\n",
			IPROC, billions, tests, errs);
		tryalone(batch_size);
		checksync();
		fflush(stderr);

		fprintf(stderr,
"PE%02x test %ld%09ld (%ld errors); Random phase...\n",
			IPROC, billions, tests, errs);
		tryrand(batch_size);
		checksync();
		fflush(stderr);

		fprintf(stderr,
"PE%02x test %ld%09ld (%ld errors); all flipping phase...\n",
			IPROC, billions, tests, errs);
		tryflips(batch_size);
		checksync();
		fflush(stderr);

		fprintf(stderr,
"PE%02x test %ld%09ld (%ld errors); Random phase...\n",
			IPROC, billions, tests, errs);
		tryrand(batch_size);
		checksync();
		fflush(stderr);
	}

	/* We'll never get here! */
	p_exit();
	return(0);
}
#else
int
main(register int argc,
register char **argv)
{
	fprintf(stderr, "Currently, ttltester doesn't support NPROC = 1\n");
	exit(1);
}
#endif
